import os
import logging
from nipype.pipeline import engine as pe
from nipype.interfaces import fsl
from nipype.interfaces.dcmstack import CopyMeta
from nipype.interfaces.utility import (IdentityInterface, Function, Merge)
import qiutil
from qiutil.logging import logger
from .. import CONF_DIR
from ..interfaces import (Gate, XNATUpload, XNATFind)
from ..helpers.bolus_arrival import bolus_arrival_index, BolusArrivalError
from .workflow_base import WorkflowBase
from .pipeline_error import PipelineError

PK_PREFIX = 'pk'
"""The XNAT modeling resource object label prefix."""

TECHNIQUES = ['mock']
"""The built-in modeling techniques."""

FIXED_R1_0_OUTPUTS = ['r1_series', 'pk_params', 'fxr_k_trans', 'fxr_v_e',
           'fxr_tau_i', 'fxr_chisq', 'fxl_k_trans', 'fxl_v_e',
           'fxl_chisq', 'delta_k_trans']
"""The modeling outputs for all runs."""

INFERRED_R1_0_OUTPUTS = FIXED_R1_0_OUTPUTS + ['dce_baseline', 'r1_0']
"""The inferred R1_0 modeling outputs."""

FASTFIT_PARAMS_FILE = 'fastfit.csv'
"""The fastfit parameters CSV file name."""

FASTFIT_PARAMS_TEMPLATE = os.path.join(CONF_DIR, FASTFIT_PARAMS_FILE)
"""The fastfit parameters template file location."""

MODELING_CONF_FILE = os.path.join(CONF_DIR, 'modeling.cfg')
"""The modeling workflow configuration."""

MODELING_PROFILE_FILE = 'profile.cfg'
"""The modeling profile file name."""


class ModelingError(Exception):
    pass


def run(technique, project, subject, session, scan, time_series,
        **opts):
    """
    Creates a :class:`qipipe.pipeline.modeling.ModelingWorkflow` and
    runs it on the given inputs.

    :param technique: the :attr:ModelingWorkflow.technique`
    :param project: the project name
    :param subject: the input subject
    :param session: the input session
    :param scan: input scan
    :param time_series: the input 4D NIfTI time series
    :param opts: the :class:`qipipe.pipeline.modeling.ModelingWorkflow`
        initializer and run options
    :return: the :meth:`qipipe.pipeline.modeling.ModelingWorkflow.run`
        result
    """
    run_opts = {key: opts.pop(key) for key in ['mask', 'registration']
                if key in opts}
    mask = opts.pop('mask', None)
    wf = ModelingWorkflow(project=project, technique=technique, **opts)

    return wf.run(subject, session, scan, time_series, **run_opts)


def generate_resource_name(technique):
    """
    Makes a unique modeling resource name. Uniqueness permits more than
    one resource to be stored for a given session without a name conflict.

    :param technique: the modeling technique
    :return: a unique XNAT modeling resource name
    """
    return "%s_%s_%s" % (PK_PREFIX, technique,
                         qiutil.file.generate_file_name())


class ModelingWorkflow(WorkflowBase):
    """
    The ModelingWorkflow builds and executes the Nipype pharmacokinetic
    mapping workflow.

    The workflow calculates the modeling parameters for an input 4D
    time series NIfTI image file as follows:

    - Compute the |R10| value, if it is not given in the options

    - Convert the DCE time series to a R1 map series

    - Determine the AIF and R1 fit parameters from the time series

    - Optimize the BOLERO pharmacokinetic model

    - Upload the modeling result to XNAT

    The modeling workflow input is the `input_spec` node consisting of the
    following input fields:

    - *subject*: the subject name

    - *session*: the session name

    - *mask*: the mask to apply to the images

    - *time_series*: the 4D time series NIfTI file to model

    - *bolus_arrival_index*: the bolus uptake volume index

    - the R1 modeling parameters described below

    If an input field is defined in the configuration file ``R1``
    topic, then the input field is set to that value.

    If the |R10| option is not set, then it is computed from the proton
    density weighted scans and DCE series baseline image.

    The outputs are collected in the `output_spec` node for the FXL
    (`Tofts standard`_) model and the FXR (`shutter speed`_) model with
    the following fields:

    - `r1_series`: the R1 series files

    - `pk_params`: the AIF and R1 parameter CSV file

    - `fxr_k_trans`, `fxl_k_trans`: the |Ktrans| vascular permeability
       transfer constant

    - `delta_k_trans`: the FXR-FXL |Ktrans| difference

    - `fxr_v_e`, `fxl_v_e`: the |ve| extravascular extracellular volume
       fraction

    - `fxr_tau_i`: the |taui| intracellular |H2O| mean lifetime

    - `fxr_chi_sq`, `fxl_chi_sq`: the |chisq| intensity goodness of fit

    In addition, if |R10| is computed, then the output includes the
    following fields:

    - `pdw_image`: the proton density weighted image

    - `dce_baseline`: the DCE series baseline image

    - `r1_0`: the computed |R10| value

    This workflow is adapted from the `AIRC DCE`_ implementation.

    :Note: This workflow uses proprietary OHSU AIRC software, notably the
        BOLERO implementation of the shutter speed model.

    .. reST substitutions:
    .. include:: <isogrk3.txt>
    .. |H2O| replace:: H\ :sub:`2`\ O
    .. |Ktrans| replace:: K\ :sup:`trans`
    .. |ve| replace:: v\ :sub:`e`
    .. |taui| replace:: |tau|\ :sub:`i`
    .. |chisq| replace:: |chi|\ :sup:`2`
    .. |R10| replace:: R1\ :sub:`0`

    .. _Tofts standard: http://onlinelibrary.wiley.com/doi/10.1002/(SICI)1522-2586(199909)10:3%3C223::AID-JMRI2%3E3.0.CO;2-S/abstract
    .. _shutter speed: http://www.ncbi.nlm.nih.gov/pmc/articles/PMC2582583
    .. _AIRC DCE: https://everett.ohsu.edu/hg/qin_dce
    """

    def __init__(self, **kwargs):
        """
        The modeling parameters can be defined in either the options or the
        configuration as follows:

        - The parameters can be defined in the configuration ``R1``
          section.

        - The keyword arguments take precedence over the configuration
          settings.

        - The *r1_0_val* takes precedence over the R1_0 computation
          fields *pd_dir* and *max_r1_0*. If *r1_0_val* is set
          in the input options, then *pd_dir* and *max_r1_0* are
          not included from the result.

        - If *pd_dir* and *max_r1_0* are set in the input options
          and *r1_0_val* is not set in the input options, then
          a *r1_0_val* configuration setting is ignored.

        - The *baseline_end_idx* defaults to 1 if it is not set in
          either the input options or the configuration.

        :param kwargs: the :class:`qipipe.pipeline.workflow_base.WorkflowBase`
            initializer keyword arguments, as well as the following options:
        :keyword resource: the XNAT resource name
        :keyword r1_0_val: the optional fixed |R10| value
        :keyword max_r1_0: the maximum computed |R10| value, if the fixed
            |R10| option is not set
        :keyword pd_dir: the proton density files parent directory, if the
            fixed |R10| option is not set
        :keyword baseline_end_idx: the number of volumes to merge into a R1
            series baseline image (default is 1)
        """
        super(ModelingWorkflow, self).__init__(logger=logger(__name__),
                                               **kwargs)

        technique_opt = kwargs.pop('technique', None)
        if not technique_opt:
            raise PipelineError('The modeling technique was not specified.')
        self.technique = technique_opt.lower()
        """The modeling technique. Built-in techniques include ``mock``."""

        rsc_opt = kwargs.pop('resource', None)
        self.resource = rsc_opt or generate_resource_name(self.technique)
        """
        The XNAT resource name for all executions of this
        :class:`qipipe.pipeline.modeling.ModelingWorkflow` instance.
        The name is unique, which permits more than one model to be
        stored for each input volume without a name conflict.
        """

        self.workflow = self._create_workflow(**kwargs)
        """
        The modeling workflow described in
        :class:`qipipe.pipeline.modeling.ModelingWorkflow`.
        """

    def run(self, subject, session, scan, time_series, **opts):
        """
        Executes the modeling workflow described in
        :class:`qipipe.pipeline.modeling.ModelingWorkflow`
        on the given input time series resource. The time series can
        be the merged scan NIFTI files or merged registration files.

        This run method connects the given inputs to the modeling workflow
        inputs. The execution workflow is then executed, resulting in a
        new uploaded XNAT resource.

        :param subject: the subject name
        :param session: the session name
        :param scan: the scan number
        :param time_series: the 4D time series resource name
        :param opts: the following keyword parameters:
        :option mask: the XNAT mask resource name
        :option registration: the XNAT registration resource name
        :return: the modeling result XNAT resource name
        """
        self.logger.debug("Modeling the %s %s Scan %d time series %s..." %
            (subject, session, scan, time_series))

        # Determine the bolus uptake. If it could not be determined,
        # then take the first volume as the uptake.
        try:
            bolus_arv_ndx = bolus_arrival_index(time_series)
        except BolusArrivalError:
            bolus_arv_ndx = 0

        # The keyword parameters.
        mask = opts.get('mask')
        registration = opts.get('registration')

        # Set the workflow input.
        input_spec = self.workflow.get_node('input_spec')
        input_spec.inputs.subject = subject
        input_spec.inputs.session = session
        input_spec.inputs.scan = scan
        input_spec.inputs.time_series = time_series
        input_spec.inputs.bolus_arrival_index = bolus_arv_ndx
        if mask:
            input_spec.inputs.mask = mask
        if registration:
            input_spec.inputs.registration = registration

        # Execute the modeling workflow.
        self.logger.debug("Executing the %s workflow on the %s %s scan %d"
                          " time series %s..." %
                          (self.workflow.name, subject, session, scan, time_series))
        self._run_workflow(self.workflow)
        self.logger.debug("Executed the %s workflow on the %s %s scan %d"
                          " time series %s." %
                          (self.workflow.name, subject, session, scan, time_series))

        # Return the resource name.
        return self.resource

    def _create_workflow(self, **opts):
        """
        Builds the modeling workflow.

        :param opts: the additional workflow initialization parameters
        :return: the Nipype workflow
        """
        self.logger.debug("Building the modeling workflow...")

        # The supervisory workflow.
        mdl_wf = pe.Workflow(name='modeling', base_dir=self.base_dir)

        # The default modeling technique is the OHSU proprietary bolero.
        #
        # TODO - generalize workflow techniques here and in registration
        # to a module reference, e.g.:
        #
        # qipipe.cfg:
        # [Modeling]
        # technique = ohsu.modeling.bolero
        #
        # New git project with:
        # qipipe-ohsu/
        #   requirements.txt:
        #     qipipe==x.x.x
        #   modeling/bolero.py:
        #     def create_workflow(**opts):
        #         ...
        #
        # pip install -e git+...
        #
        # Then replace below with:
        #
        # modules = technique_opt.split('.')
        # parent_opt = '.'.join(modules[:-1])
        # child_opt = modules[-1]
        # if parent_opt:
        #     wf_gen = __import__(parent_opt, globals(), locals(), [child_opt])
        # else:
        #     wf_gen = __import__(child_opt)
        # base_wf = wf_gen.create_workflow(**opts)
        #
        if self.technique == 'bolero':
            base_wf = self._create_bolero_workflow(**opts)
        elif self.technique == 'mock':
            base_wf = self._create_mock_workflow(**opts)
        elif self.technique:
            raise ModelingError("The modeling technique is unsupported:"
                                " %s" % self.technique)
        else:
            raise ModelingError('The modeling technique is missing')

        # The workflow input fields.
        in_fields = ['subject', 'session', 'scan', 'time_series', 'mask',
                     'bolus_arrival_index']
        input_xfc = IdentityInterface(fields=in_fields)
        # The profile location is a temp file.
        input_spec = pe.Node(input_xfc, name='input_spec')
        self.logger.debug("The modeling workflow input is %s with"
            " fields %s" % (input_spec.name, in_fields))
        mdl_wf.connect(input_spec, 'time_series',
                       base_wf, 'input_spec.time_series')
        mdl_wf.connect(input_spec, 'mask', base_wf, 'input_spec.mask')
        mdl_wf.connect(input_spec, 'bolus_arrival_index',
                       base_wf, 'input_spec.bolus_arrival_index')

        # Upload the modeling results to XNAT.
        # Each output field contains a modeling result file.
        # Upload these files to the modeling resource.
        base_output = base_wf.get_node('output_spec')
        out_fields = base_output.outputs.copyable_trait_names()
        merge_outputs = pe.Node(Merge(len(out_fields)),
                                name='merge_outputs')
        for i, field in enumerate(out_fields):
            base_field = 'output_spec.' + field
            mdl_wf.connect(base_wf, base_field,
                           merge_outputs, "in%d" % (i + 1))

        # Make the resource.
        cr_rsc_xfc = XNATFind(project=self.project, resource=self.resource,
                              modality='MR', create=True)
        create_resource = pe.Node(cr_rsc_xfc, name='create_resource')
        mdl_wf.connect(input_spec, 'subject', create_resource, 'subject')
        mdl_wf.connect(input_spec, 'session', create_resource, 'session')
        mdl_wf.connect(input_spec, 'scan', create_resource, 'scan')

        # Gate uploads on the create_resource node.
        rsc_gate_xfc = Gate(fields=['resource', 'xnat_id'], resource=self.resource)
        resource_gate = pe.Node(rsc_gate_xfc, name='resource_gate')
        # xnat_id is not subsequently used. It is a dead-end connection
        # whose sole purpose is to gate successor nodes on create_resource.
        mdl_wf.connect(create_resource, 'xnat_id', resource_gate, 'xnat_id')

        # Upload the outputs.
        upload_xfc = XNATUpload(project=self.project)
        upload_outputs = pe.Node(upload_xfc, name='upload_outputs')
        mdl_wf.connect(input_spec, 'subject', upload_outputs, 'subject')
        mdl_wf.connect(input_spec, 'session', upload_outputs, 'session')
        mdl_wf.connect(input_spec, 'scan', upload_outputs, 'scan')
        mdl_wf.connect(resource_gate, 'resource', upload_outputs, 'resource')
        mdl_wf.connect(merge_outputs, 'out', upload_outputs, 'in_files')

        # Make a gate whose sole purpose is to tie the input_spec node
        # to create_profile.
        cr_prf_gate_xfc = Gate(fields=['scan', 'technique'], technique=self.technique)
        create_profile_gate = pe.Node(cr_prf_gate_xfc, name='create_profile_gate')
        # scan is not subsequently used. It is a dead-end connection
        # whose sole purpose is to gate successor nodes on create_profile.
        mdl_wf.connect(input_spec, 'scan', create_profile_gate, 'scan')

        # Make the profile.
        create_profile_xfc = Function(input_names=['technique'],
                                      output_names=['out_file'],
                                      function=create_profile)
        create_profile_func = pe.Node(create_profile_xfc, name='create_profile')
        mdl_wf.connect(create_profile_gate, 'technique',
                       create_profile_func, 'technique')

        # Upload the profile.
        upload_profile = pe.Node(upload_xfc, name='upload_profile')
        mdl_wf.connect(input_spec, 'subject', upload_profile, 'subject')
        mdl_wf.connect(input_spec, 'session', upload_profile, 'session')
        mdl_wf.connect(input_spec, 'scan', upload_profile, 'scan')
        mdl_wf.connect(resource_gate, 'resource', upload_profile, 'resource')
        mdl_wf.connect(create_profile_func, 'out_file',
                       upload_profile, 'in_files')

        # TODO - Get the overall and ROI FSL mean intensity values.

        # The output is the modeling outputs.
        output_xfc = IdentityInterface(fields=out_fields)
        output_spec = pe.Node(output_xfc, name='output_spec')
        for field in out_fields:
            base_field = 'output_spec.' + field
            mdl_wf.connect(base_wf, base_field, output_spec, field)
        self.logger.debug("The modeling workflow output is %s with"
                           " fields %s" % (output_spec.name, out_fields))

        self._configure_nodes(mdl_wf)

        self.logger.debug("Created the %s workflow." % mdl_wf.name)
        # If debug is set, then diagram the workflow graph.
        if self.logger.level <= logging.DEBUG:
            self.depict_workflow(mdl_wf)

        return mdl_wf

    def _create_bolero_workflow(self, **opts):
        """
        Creates the modeling base workflow. This workflow performs the steps
        described in :class:`qipipe.pipeline.modeling.ModelingWorkflow` with
        the exception of XNAT upload.

        :Note: This workflow is adapted from the AIRC workflow at
        https://everett.ohsu.edu/hg/qin_dce. The AIRC workflow time series
        merge is removed and added as input to the workflow created by this
        method. Any change to the ``qin_dce`` workflow should be reflected in
        this method.

        :param opts: the PK modeling parameters
        :return: the pyxnat Workflow
        """
        base_wf = pe.Workflow(name='bolero', base_dir=self.base_dir)

        # The PK modeling parameters.
        opts = self._r1_parameters(**opts)
        # Set the use_fixed_r1_0 flag.
        use_fixed_r1_0 = opts.get('r1_0_val') != None

        # Set up the input node.
        non_pk_flds = ['time_series', 'mask', 'bolus_arrival_index']
        in_fields = non_pk_flds + opts.keys()
        input_xfc = IdentityInterface(fields=in_fields, **opts)
        input_spec = pe.Node(input_xfc, name='input_spec')
        # Set the config parameters.
        for field in in_fields:
            if field in opts:
                setattr(input_spec.inputs, field, opts[field])

        # If we are not using a fixed r1_0 value, then compute a map
        # from a proton density weighted scan and the baseline of the
        # DCE series.
        if not use_fixed_r1_0:
            # Create the DCE baseline image.
            make_base_func = Function(
                input_names=['time_series', 'baseline_end_idx'],
                output_names=['baseline'],
                function=make_baseline),
            make_base = pe.Node(make_base_func, name='make_base')
            base_wf.connect(input_spec, 'time_series',
                            make_base, 'time_series')
            base_wf.connect(input_spec, 'baseline_end_idx',
                            make_base, 'baseline_end_idx')

            # Create the R1_0 map.
            get_r1_0_func = Function(
                input_names=['pdw_image', 't1w_image', 'max_r1_0', 'mask'],
                output_names=['r1_0_map'],
                function=make_r1_0),
            get_r1_0 = pe.Node(get_r1_0_func, name='get_r1_0')
            base_wf.connect(input_spec, 'pd_nii', get_r1_0, 'pdw_image')
            base_wf.connect(make_base, 'baseline', get_r1_0, 't1w_image')
            base_wf.connect(input_spec, 'max_r1_0', get_r1_0, 'max_r1_0')
            base_wf.connect(input_spec, 'mask', get_r1_0, 'mask')

        # The R1 destination directory.
        # Convert the DCE time series to R1 maps.
        get_r1_series_func = Function(
            input_names=['time_series', 'r1_0', 'baseline_end', 'mask', 'dest'],
            output_names=['r1_series'], function=make_r1_series)
        get_r1_series = pe.Node(get_r1_series_func, dest=self.base_dir,
                                name='get_r1_series')
        base_wf.connect(input_spec, 'time_series',
                        get_r1_series, 'time_series')
        base_wf.connect(input_spec, 'baseline_end_idx',
                        get_r1_series, 'baseline_end')
        base_wf.connect(input_spec, 'mask',
                        get_r1_series, 'mask')
        if use_fixed_r1_0:
            base_wf.connect(input_spec, 'r1_0_val', get_r1_series, 'r1_0')
        else:
            base_wf.connect(get_r1_0, 'r1_0_map', get_r1_series, 'r1_0')

        # Copy the time series meta-data to the R1 series.
        copy_meta = pe.Node(CopyMeta(), name='copy_meta')
        copy_meta.inputs.include_classes = [('global', 'const'),
                                            ('time', 'samples')]
        base_wf.connect(input_spec, 'time_series', copy_meta, 'src_file')
        base_wf.connect(get_r1_series, 'r1_series', copy_meta, 'dest_file')

        # Get the pharmacokinetic mapping parameters.
        get_params_flds = ['time_series', 'bolus_arrival_index']
        get_params_func = Function(input_names=get_params_flds,
                                   output_names=['params_csv'],
                                   function=get_fit_params)
        get_params = pe.Node(get_params_func, name='get_params')
        base_wf.connect(input_spec, 'time_series', get_params, 'time_series')
        base_wf.connect(input_spec, 'bolus_arrival_index',
                        get_params, 'bolus_arrival_index')

        # Import Fastfit on demand. This allows the modeling module to be
        # imported by clients without the proprietary Fastfit modeling
        # tool.
        try:
            Fastfit
        except NameError:
            from ..interfaces.fastfit import Fastfit
        # Optimize the pharmacokinetic model.
        pk_map = pe.Node(Fastfit(), name='pk_map')
        pk_map.inputs.model_name = 'fxr.model'
        pk_map.inputs.optional_outs = ['chisq', 'guess_model.k_trans',
                                       'guess_model.v_e', 'guess_model.chisq']
        base_wf.connect(copy_meta, 'dest_file', pk_map, 'target_data')
        base_wf.connect(input_spec, 'mask', pk_map, 'mask')
        base_wf.connect(get_params, 'params_csv', pk_map, 'params_csv')
        # Set the MPI flag.
        pk_map.inputs.use_mpi = self.is_distributable

        # Compute the Ktrans difference.
        delta_k_trans = pe.Node(fsl.ImageMaths(), name='delta_k_trans')
        delta_k_trans.inputs.op_string = '-sub'
        base_wf.connect(pk_map, 'k_trans', delta_k_trans, 'in_file')
        base_wf.connect(pk_map, 'guess_model.k_trans',
                        delta_k_trans, 'in_file2')

        # The modeling outputs.
        outputs = FIXED_R1_0_OUTPUTS if use_fixed_r1_0 else INFERRED_R1_0_OUTPUTS
        output_spec = pe.Node(IdentityInterface(fields=outputs),
                              name='output_spec')
        # Run the PK model.
        base_wf.connect(copy_meta, 'dest_file', output_spec, 'r1_series')
        base_wf.connect(get_params, 'params_csv', output_spec, 'pk_params')
        base_wf.connect(pk_map, 'k_trans', output_spec, 'fxr_k_trans')
        base_wf.connect(pk_map, 'v_e', output_spec, 'fxr_v_e')
        base_wf.connect(pk_map, 'tau_i', output_spec, 'fxr_tau_i')
        base_wf.connect(pk_map, 'chisq', output_spec, 'fxr_chisq')
        base_wf.connect(pk_map, 'guess_model.k_trans',
                        output_spec, 'fxl_k_trans')
        base_wf.connect(pk_map, 'guess_model.v_e', output_spec, 'fxl_v_e')
        base_wf.connect(pk_map, 'guess_model.chisq', output_spec, 'fxl_chisq')
        base_wf.connect(delta_k_trans, 'out_file',
                        output_spec, 'delta_k_trans')
        # If we are inferring R1_0, then make the DCE baseline.
        if not use_fixed_r1_0:
            base_wf.connect(make_base, 'baseline',
                            output_spec, 'dce_baseline')
            base_wf.connect(get_r1_0, 'r1_0_map', output_spec, 'r1_0')

        self._configure_nodes(base_wf)

        return base_wf

    def _create_mock_workflow(self, **opts):
        """
        Creates a dummy modeling base workflow. This workflow performs
        the steps described in
        :class:`qipipe.pipeline.modeling.ModelingWorkflow` with the
        exception of XNAT upload.

        :param opts: the PK modeling parameters
        :return: the pyxnat Workflow
        """
        base_wf = pe.Workflow(name='mock', base_dir=self.base_dir)

        # The PK modeling parameters.
        opts = self._r1_parameters(**opts)

        # Set up the input node.
        non_pk_flds = ['time_series', 'mask', 'bolus_arrival_index']
        in_fields = non_pk_flds + opts.keys()
        input_xfc = IdentityInterface(fields=in_fields, **opts)
        input_spec = pe.Node(input_xfc, name='input_spec')
        # Set the config parameters.
        for field in in_fields:
            if field in opts:
                setattr(input_spec.inputs, field, opts[field])

        # Get the pharmacokinetic mapping parameters.
        get_params_flds = ['time_series', 'bolus_arrival_index']
        get_params_func = Function(input_names=get_params_flds,
                                   output_names=['params_csv'],
                                   function=get_fit_params)
        get_params = pe.Node(get_params_func, name='get_params')
        base_wf.connect(input_spec, 'time_series', get_params, 'time_series')
        base_wf.connect(input_spec, 'bolus_arrival_index',
                        get_params, 'bolus_arrival_index')

        # Set up the outputs.
        outputs = ['pk_params']
        output_spec = pe.Node(IdentityInterface(fields=outputs),
                              name='output_spec')
        base_wf.connect(get_params, 'params_csv', output_spec, 'pk_params')

        self._configure_nodes(base_wf)

        return base_wf

    def _r1_parameters(self, **opts):
        """
        Collects the R1 modeling parameters defined in either the options
        or the configuration, as described in :class:`ModelingWorkflow`.

        :param opts: the input options
        :return: the parameter {name: value} dictionary
        """
        config = self.configuration.get('R1', {})

        logger(__name__).debug("Setting the R1 parameters from the option"
            " keyword parameters %s and configuration %s..." % (opts, config))

        # The R1_0 computation fields.
        r1_fields = ['pd_dir', 'max_r1_0']
        # All of the possible fields.
        fields = set(r1_fields)
        fields.update(['baseline_end_idx', 'r1_0_val'])
        # The PK options.
        pk_opts = {k: opts[k] for k in fields if k in opts}
        if 'baseline_end_idx' not in pk_opts:
            # Look for the the baseline parameter in the configuration.
            if 'baseline_end_idx' in config:
                pk_opts['baseline_end_idx'] = config['baseline_end_idx']
            else:
                # The default baseline image count.
                pk_opts['baseline_end_idx'] = 1

        # Set the use_fixed_r1_0 variable to None, signifying unknown.
        use_fixed_r1_0 = None
        # Get the R1_0 parameter values.
        if 'r1_0_val' in pk_opts:
            r1_0_val = pk_opts.get('r1_0_val')
            if r1_0_val:
                use_fixed_r1_0 = True
            else:
                use_fixed_r1_0 = False
        else:
            for field in r1_fields:
                value = pk_opts.get(field)
                if value:
                    use_fixed_r1_0 = False

        # If none of the R1_0 options are set in the options,
        # then try the configuration.
        if use_fixed_r1_0 == None:
            r1_0_val = config.get('r1_0_val')
            if r1_0_val:
                pk_opts['r1_0_val'] = r1_0_val
                use_fixed_r1_0 = True

        # If R1_0 is not fixed, then augment the R1_0 options
        # from the configuration, if necessary.
        if not use_fixed_r1_0:
            for field in r1_fields:
                if field not in pk_opts and field in config:
                    use_fixed_r1_0 = False
                    pk_opts[field] = config[field]
                # Validate the R1 parameter.
                if not pk_opts.get(field):
                    raise ModelingError("Missing both the r1_0_val and the %s"
                        " parameter." % field)

        # If the use_fixed_r1_0 flag is set, then remove the
        # extraneous R1 computation fields.
        if use_fixed_r1_0:
            for field in r1_fields:
                pk_opts.pop(field, None)

        self.logger.debug("The PK modeling parameters: %s" % pk_opts)

        return pk_opts


### Utility functions called by workflow nodes. ###


def make_baseline(time_series, baseline_end_idx):
    """
    Makes the R1_0 computation baseline NIfTI file.

    :param time_series: the modeling input 4D NIfTI image
    :param baseline_end_idx: the exclusive limit of the baseline
        computation input series
    :return: the baseline NIfTI file name
    :raise ModelingError: if the end index is a negative number
    """
    from dcmstack.dcmmeta import NiftiWrapper

    if baseline_end_idx <= 0:
        raise ModelingError("The R1_0 computation baseline end index input value"
                            " is not a positive number: %s" % baseline_end_idx)
    nii = nb.load(time_series)
    nw = NiftiWrapper(nii)

    baselines = []
    for idx, split_nii in nw.split():
        if idx == baseline_end_idx:
            break
        baselines.append(split_nii)

    if len(baselines) == 1:
        return baselines[0]
    else:
        baseline = NiftiWrapper.from_sequence(baselines)
    baseline_path = path.join(os.getcwd(), 'baseline.nii.gz')
    nb.save(baseline, baseline_path)

    return baseline_path


def make_r1_0(pdw_image, t1w_image, max_r1_0, **kwargs):
    """
    Returns the R1_0 map NIfTI file from the given proton density
    and T1-weighted images. The R1_0 map is computed using the
    ``pdw_t1w_to_r1`` function. The ``pdw_t1w_to_r1`` module
    must be in the Python path.

    :param pdw_image: the proton density NIfTI image file path
    :param t1w_image: the T1-weighted image file path
    :param max_r1_0: the R1_0 range maximum
    :param kwargs: the ``pdw_t1w_to_r1`` function keyword arguments
    :return: the R1_0 map NIfTI image file name
    """
    import os
    import nibabel as nb
    import numpy as np
    from pdw_t1w_to_r1 import pdw_t1w_to_r1
    from dcmstack.dcmmeta import NiftiWrapper

    pdw_nw = NiftiWrapper(nb.load(pdw_image), make_empty=True)
    t1w_nw = NiftiWrapper(nb.load(t1w_image), make_empty=True)
    r1_space = np.arange(0.01, max_r1_0, 0.01)
    mask_opt = kwargs.pop('mask', None)
    if mask_opt:
        kwargs['mask'] = nb.load(mask_opt).get_data()
    r1_0 = pdw_t1w_to_r1(pdw_nw, t1w_nw, r1_space=r1_space, **kwargs)

    cwd = os.getcwd()
    out_nii = nb.Nifti1Image(r1_0, pdw_nw.nii_img.get_affine())
    out_fn = os.path.join(cwd, 'r1_0_map.nii.gz')
    nb.save(out_nii, out_fn)

    return out_fn


def make_r1_series(time_series, r1_0, **kwargs):
    """
    Makes the R1_0 series NIfTI file.

    :param time_series: the modeling input 4D NIfTI image
    :param r1_0: the R1_0 image file path
    :param kwargs: the ``dce_to_r1`` keyword options
    :return: the R1_0 series NIfTI image file name
    """
    import os
    import nibabel as nb
    from dce_to_r1 import dce_to_r1
    from dcmstack.dcmmeta import NiftiWrapper

    dce_nw = NiftiWrapper(nb.load(time_series), make_empty=True)
    if not isinstance(r1_0, float):
        r1_0 = nb.load(r1_0).get_data()
    mask_opt = kwargs.pop('mask', None)
    if mask_opt:
        kwargs['mask'] = nb.load(mask_opt).get_data()
    r1_series = dce_to_r1(dce_nw, r1_0, **kwargs)

    cwd = os.getcwd()
    out_nii = nb.Nifti1Image(r1_series, dce_nw.nii_img.get_affine())
    out_file = os.path.join(cwd, 'r1_series.nii.gz')
    nb.save(out_nii, out_file)

    return out_file


def get_fit_params(time_series, bolus_arrival_index):
    """
    Obtains the following modeling fit parameters:

    * *aif_params*: arterial input function parameter array
    * *aif_delta_t*: acquisition time deltas
    * *aif_shift*: acquisition time shift
    * *r1_cr*: contrast R1
    * *r1_b_pre*: pre-contrast R1

    :param time_series: the modeling input 4D NIfTI image
    :param bolus_arrival_index: the bolus uptake series index
    :return: the parameter CSV file path
    """
    import os
    import csv
    import nibabel as nb
    import numpy as np
    from dcmstack.dcmmeta import NiftiWrapper
    from dcmstack import dcm_time_to_sec
    from qipipe.pipeline.modeling import (FASTFIT_PARAMS_TEMPLATE,
                                          FASTFIT_PARAMS_FILE)

    # Load the time series into a NIfTI wrapper.
    nii = nb.load(time_series)
    nw = NiftiWrapper(nii)

    # The AIF shift parameter is the offset of the bolus arrival
    # series mid-point acquisition time from the MR session start
    # time.
    acq_time0 = dcm_time_to_sec(nw.get_meta('AcquisitionTime', (0, 0, 0, 0)))
    acq_time1 = dcm_time_to_sec(
        nw.get_meta('AcquisitionTime', (0, 0, 0, bolus_arrival_index)))
    acq_time2 = dcm_time_to_sec(
        nw.get_meta('AcquisitionTime', (0, 0, 0, bolus_arrival_index + 1)))
    aif_shift = ((acq_time1 + acq_time2) / 2.0) - acq_time0

    # The static input parameter CSV template.
    with open(FASTFIT_PARAMS_TEMPLATE) as csv_file:
        rows = list(csv.reader(csv_file))
    # Add the shift.
    rows.append(['aif_shift', str(aif_shift)])

    # Create the parameter CSV file.
    with open(FASTFIT_PARAMS_FILE, 'w') as csv_file:
        csv_writer = csv.writer(csv_file)
        csv_writer.writerows(rows)

    return os.path.join(os.getcwd(), FASTFIT_PARAMS_FILE)


def create_profile(technique, dest_file=None):
    """
    Creates the modeling profile CSV file from the
    :const:`MODELING_CONF_FILE` ``R1`` topic.

    :param dest_file: the target profile location
        (default :const:`MODELING_PROFILE_FILE` in the current directory)
    :return: the destination file
    """
    import os
    import csv
    from qiutil.ast_config import read_config
    from qipipe.pipeline.modeling import (MODELING_CONF_FILE,
                                          MODELING_PROFILE_FILE,
                                          ModelingError)

    # Make the R1 params file.
    cfg = read_config(MODELING_CONF_FILE)
    cfg_dict = dict(cfg)
    r1_opts = cfg_dict.get('R1')
    if not r1_opts:
        raise ModelingError("The imaging configuration file %s"
                            " is missing the R1 topic" %
                            configuration)
    if not dest_file:
        dest_file = os.path.join(os.getcwd(), MODELING_PROFILE_FILE)
    with open(dest_file, 'w+') as csv_file:
        csv_writer = csv.writer(csv_file)
        csv_writer.writerow(('technique', technique))
        csv_writer.writerows(r1_opts.items())

    return dest_file




"""
This module updates the qiprofile database Subject pathology information
from the pathology Excel workbook file.
"""
import six
from qiprofile_rest_client.model.clinical import (
    SarcomaPathology, FNCLCCGrade, NecrosisPercentValue,
    NecrosisPercentRange, necrosis_percent_as_score
)
from . import parse
from .pathology import (PathologyError, PathologyWorksheet, PathologyUpdate)

COL_ATTRS = {'Tumor Location': 'location'}
"""
The following special column: attribute associations:
* The ``Tumor Location`` column corresponds to the pathology ``location`` 
  attribute
"""

PARSERS = dict(necrosis_percent=lambda s: _parse_necrosis_percent(s))
"""
The following special parsers:
* The necrosis percent can be an integer or a range, e.g. ``80-90``.
"""


def read(workbook, **condition):
    """
    This is a convenience method that wraps :class:`SarcomaPathologyWorksheet`
    :meth:`qipipe.qiprofile.xls.Worksheet.read`.

    :param workbook: the read-only ``openpyxl`` workbook object
    :param condition: the :meth:`qipipe.qiprofile.xls.Worksheet.read`
        filter condition
    :return: the :meth:`qipipe.qiprofile.xls.Worksheet.read` rows
    """
    reader = PathologyWorksheet(workbook, SarcomaPathology, FNCLCCGrade,
                                parsers=dict(PARSERS),
                                column_attributes=COL_ATTRS)
    
    return reader.read(**condition)


def update(subject, rows):
    """
    Updates the given subject data object from the Sarcoma pathology XLS
    rows.

    :param subject: the ``Subject`` Mongo Engine database object
        to update
    :param rows: the input pathology :meth:`read` rows list 
    """
    updater = SarcomaPathologyUpdate(subject)
    updater.update(rows)


def _parse_necrosis_percent(value):
    """
    :param value: the input XLS value
    :return: the necrosis percent database object
    :raise qipipe.qiprofile.parse.ParseError: if the value cannot be parsed
    """
    if isinstance(value, int):
        return NecrosisPercentValue(value=value)
    elif not value:
        return None
    elif not isinstance(value, six.string_types):
        raise parse.ParseError("The input necrosis percent type is not"
                         " supported: %s (%s)" % (value, value.__class__))
    
    # Parse the string input.
    values = [int(bound) for bound in value.split('-')]
    if len(values) == 1:
        return NecrosisPercentValue(value=values[0])
    else:
        start_val, stop_val = values
        start_bnd = NecrosisPercentRange.LowerBound(value=start_val)
        stop_bnd = NecrosisPercentRange.UpperBound(value=stop_val)
        return NecrosisPercentRange(start=start_bnd, stop=stop_bnd)


class SarcomaPathologyUpdate(PathologyUpdate):
    """The Sarcoma pathology update facade."""
    
    def __init__(self, subject):
        """
        :param subject: the ``Subject`` Mongo Engine database object
            to update
        """
        super(SarcomaPathologyUpdate, self).__init__(
            subject, tumor_type='Sarcoma', pathology_class=SarcomaPathology,
            grade_class=FNCLCCGrade
        )

    def pathology_content(self, row):
        """
        Collects the pathology object from the given input row.
        This subclass implementation adds the following items:
        * If there are *necrosis_percent* and *tnm* items, then the TNM
          necrosis_score is inferred from the necrosis percent
        
        :param row: the input row
        :return: the {attribute: value} content dictionary
        """
        # Delegate to PathologyUpdate for the common pathology fields.
        content = super(SarcomaPathologyUpdate, self).pathology_content(row)
        # If there is a necrosis percent, then calculate the necrosis score
        # from the necrosis percent.
        if 'necrosis_percent' in content and 'tnm' in content:
            necrosis_pct = content['necrosis_percent']
            tnm = content['tnm']
            tnm.necrosis_score = necrosis_percent_as_score(necrosis_pct)

        return content



import os
import re
import tempfile
import pprint
import networkx as nx
import qixnat
from qiutil.logging import logger
from qiutil.collections import EMPTY_DICT
from qiutil.ast_config import read_config
from .. import CONF_DIR
from ..helpers.distributable import DISTRIBUTABLE
from .pipeline_error import PipelineError


class WorkflowBase(object):
    """
    The WorkflowBase class is the base class for the qipipe workflow
    wrapper classes.

    If the :attr:`distributable' flag is set, then the execution is
    distributed using the Nipype plug-in specified in the configuration
    *plug_in* parameter.

    The workflow plug-in arguments and node inputs can be specified in
    a :class:`qiutil.ast_config.ASTConfig` file. The configuration
    directory order consist of the order consist of the search locations
    in low-to-high precedence order consist of the following:

    1. the qipipe module ``conf`` directory

    2. the *config_dir* initialization keyword option

    The common configuration is loaded from the ``default.cfg`` file or
    in the directory locations. The workflow-specific configuration file
    name is the lower-case name of the ``WorkflowBase`` subclass with
    ``.cfg`` extension, e.g. ``registration.cfg`` for
    :class:`qipipe.workflow.registration.RegistrationWorkflow`.
    The configuration settings are then loaded from the common configuration
    files followed by the workflow-specific configuration files.
    """

    CLASS_NAME_PAT = re.compile("^(\w+)Workflow$")
    """The workflow wrapper class name matcher."""

    INTERFACE_PREFIX_PAT = re.compile('(\w+\.)+interfaces?\.?')
    """
    Regexp matcher for an interface module.

    Example:

    >>> from qipipe.pipeline.workflow_base import WorkflowBase
    >>> WorkflowBase.INTERFACE_PREFIX_PAT.match('nipype.interfaces.ants.util.AverageImages').groups()
    ('nipype.',)
    """

    MODULE_PREFIX_PAT = re.compile('^((\w+\.)*)(\w+\.)(\w+)$')
    """
    Regexp matcher for a module prefix.

    Example:

    >>> from qipipe.pipeline.workflow_base import WorkflowBase
    >>> WorkflowBase.MODULE_PREFIX_PAT.match('ants.util.AverageImages').groups()
    ('ants.', 'ants.', 'util.', 'AverageImages')
    >>> WorkflowBase.MODULE_PREFIX_PAT.match('AverageImages')
    None
    """

    def __init__(self, **kwargs):
        """
        Initializes this workflow wrapper object.

        :param kwargs: the following keyword arguments:
        :keyword project: the :attr:`project`
        :keyword logger: the :attr:`logger` to use
        :keyword parent: the parent workflow for a child workflow
        :keyword base_dir: the :attr:`base_dir`
        :keyword config: the optional workflow node :attr:`configuration`
            file location or dictionary
        :keyword dry_run: the :attr:`dry_run` flag
        :keyword distributable: the :attr:`distributable` flag
        :raise PipelineError: if there is neither a *project* nor
            a *parent* argument
        """
        parent = kwargs.get('parent')
        project_opt = kwargs.get('project')
        if project_opt:
            project = project_opt
        elif parent:
            project = parent.project
        else:
            raise PipelineError('The workflow wrapper must have either a'
                                ' project or a parent')
        self.project = project
        """The XNAT project name."""

        logger_opt = kwargs.get('logger')
        if logger_opt:
            _logger = logger_opt
        elif parent:
            _logger = parent.logger
        else:
            _logger = logger('qiprofile')
        self.logger = _logger
        """This workflow's logger."""

        base_dir_opt = kwargs.get('base_dir')
        if base_dir_opt:
            base_dir = base_dir_opt
        elif parent:
            base_dir = parent.base_dir
        else:
            base_dir = tempfile.mkdtemp(prefix='qipipe_')
        self.base_dir = base_dir
        "The workflow execution directory (default a new temp directory)."

        cfg_dir_opt = kwargs.get('config_dir')
        if cfg_dir_opt:
            cfg_dir = cfg_dir_opt
        elif parent:
            cfg_dir = parent.config_dir
        else:
            cfg_dir = None
        self.config_dir = cfg_dir
        """The workflow node inputs configuration directory."""

        self.configuration = self._load_configuration()
        """The workflow node inputs configuration."""

        config_s = pprint.pformat(self.configuration)
        self.logger.debug("Pipeline configuration:")
        for line in config_s.split("\n"):
            self.logger.debug(line)

        # The distributable option is only set if the qipipe command
        # option --no-submit is set. In that case, distributable is
        # set to False. Otherwise, the distributable option is not
        # set. Thus, the is_distributable attribute is set to False if
        # either the --no-submit command option was set or there is
        # no cluster environment. Otherwise, there is a cluster
        # environment and the is_distributable flag is set to True.
        distributable_opt = kwargs.get('distributable')
        if distributable_opt:
            distributable = distributable_opt
        elif parent:
            distributable = parent.is_distributable
        else:
            distributable = DISTRIBUTABLE
        self.is_distributable = distributable
        """Flag indicating whether to submit jobs to a cluster."""

        # The execution plug-in.
        if 'Execution' in self.configuration:
            exec_opts = self.configuration['Execution']
            self.plug_in = exec_opts.pop('plug-in', None)
        else:
            self.plug_in = None

        if 'dry_run' in kwargs:
            dry_run = kwargs.get('dry_run')
        elif parent:
            dry_run = parent.dry_run
        else:
            dry_run = False
        self.dry_run = dry_run
        """Flag indicating whether to prepare but not run the workflow."""

    def depict_workflow(self, workflow):
        """
        Diagrams the given workflow graph. The diagram is written to the
        *name*\ ``.dot.png`` in the workflow base directory.

        :param workflow the workflow to diagram
        """
        base = workflow.name + '.dot'
        fname = os.path.join(self.base_dir, base)
        workflow.write_graph(dotfilename=fname)
        self.logger.debug("The %s workflow graph is depicted at %s.png." %
                         (workflow.name, fname))

    def _load_configuration(self):
        """
        Loads the workflow configuration, as described in
        :class:`WorkflowBase`.

        :return: the configuration dictionary
        """
        # The configuration files to load.
        cfg_files = []
        # Validate that the workflow name ends in Workflow.
        match = WorkflowBase.CLASS_NAME_PAT.match(self.__class__.__name__)
        if not match:
            raise NameError("The workflow wrapper class does not match the"
                            " standard workflow class name pattern: %s" %
                            self.__class__.__name__)
        name = match.group(1)
        
        # The default configuration files.
        def_cfg_files = self._configuration_files('default')
        # The workflow configuration files.
        wf_cfg_files = self._configuration_files(name)
        # All path configuration files.
        cfg_files = def_cfg_files + wf_cfg_files

        # Load the configuration.
        self.logger.debug("Loading the %s configuration files %s..." %
                          (name, cfg_files))
        cfg = read_config(*cfg_files)

        return dict(cfg)

    def _configuration_files(self, name):
        """
        :param name: the case-insensitive configuration file name
            without path or extension
        :return: the list of matching configuration files in the configuration
            path, in low-to-high precedence order
        """
        
        # The file name is lower-case with .cfg extension.
        fname = name.lower() + '.cfg'
        # The configuration directories.
        cfg_dirs = [CONF_DIR]
        if self.config_dir:
            cfg_dirs.append(self.config_dir)
        # The configuration file locations.
        cfg_files = []
        for cfg_dir in cfg_dirs:
            cfg_file = os.path.join(cfg_dir, fname)
            if os.path.exists(cfg_file):
                cfg_files.append(os.path.abspath(cfg_file))

        return cfg_files


    def _download_scans(self, xnat, subject, session, dest):
        """
        Download the NIFTI scan files for the given session.

        :param xnat: the :class:`qixnat.facade.XNAT` connection
        :param subject: the XNAT subject label
        :param session: the XNAT session label
        :param dest: the destination directory path
        :return: the download file paths
        """
        return xnat.download(self.project, subject, session, dest=dest)

    def _run_workflow(self, workflow):
        """
        Executes the given workflow.

        :param workflow: the workflow to run
        """
        # If the workflow can be distributed, then get the plugin
        # arguments.
        is_dist_clause = 'is' if self.is_distributable else 'is not'
        self.logger.debug("The %s workflow %s distributable in a"
                           " cluster environment." %
                           (workflow.name, is_dist_clause))
        if self.is_distributable:
            opts = self._configure_plugin(workflow)
        else:
            opts = {}

        # Set the base directory to an absolute path.
        if workflow.base_dir:
            workflow.base_dir = os.path.abspath(workflow.base_dir)
        else:
            workflow.base_dir = self.base_dir

        # Run the workflow.
        self.logger.debug("Executing the %s workflow in %s..." %
                           (workflow.name, workflow.base_dir))
        if self.dry_run:
            self.logger.debug("Skipped workflow %s job submission,"
                               " since the dry run flag is set." %
                               workflow.name)
        else:
            with qixnat.connect():
                workflow.run(**opts)

    def _inspect_workflow_inputs(self, workflow):
        """
        Collects the given workflow nodes' inputs for debugging.

        :return: a {node name: parameters} dictionary, where *parameters*
            is a node parameter {name: value} dictionary
        """
        return {node_name: self._inspect_node_inputs(workflow.get_node(node_name))
                for node_name in workflow.list_node_names()}

    def _inspect_node_inputs(self, node):
        """
        Collects the given node inputs and plugin arguments for debugging.

        :return: the node parameter {name: value} dictionary
        """
        fields = node.inputs.copyable_trait_names()
        param_dict = {}
        for field in fields:
            value = getattr(node.inputs, field)
            if value != None:
                param_dict[field] = value

        return param_dict

    def _configure_plugin(self, workflow):
        """
        Sets the *execution* and plug-in parameters for the given workflow.
        See the ``conf`` directory files for examples.

        :param workflow: the workflow to run
        :return: the workflow execution arguments
        """
        # The execution setting.
        if 'Execution' in self.configuration:
            workflow.config['execution'] = self.configuration['Execution']
            self.logger.debug("Workflow %s execution parameters: %s." %
                             (workflow.name, workflow.config['execution']))

        # The Nipype plug-in parameters.
        if self.plug_in and self.plug_in in self.configuration:
            plug_in_opts = self.configuration[self.plug_in]
            opts = dict(plugin=self.plug_in, **plug_in_opts)
            self.logger.debug("Workflow %s %s plug-in parameters: %s." %
                             (workflow.name, self.plug_in, opts))
        else:
            opts = {}

        return opts

    def _configure_nodes(self, workflow):
        """
        Sets the input parameters defined for the given workflow in
        this WorkflowBase's configuration. This method is called by
        each WorkflowBase subclass after the workflow is built and
        prior to execution.

        :Note: nested workflow nodes are not configured, e.g. if the
            ``registration`` workflow connects a `realign`` workflow
            node ``fnirt``, then the nested ``realign.fnirt`` node in
            ``registration`` is not configured by the parent workflow.
            The nested workflow is configured separately when the nested
            WorkflowBase subclass object is created.

        :param workflow: the workflow containing the nodes
        """
        # The default plug-in setting.
        if self.is_distributable and self.plug_in and self.plug_in in self.configuration:
            plugin_cfg = self.configuration[self.plug_in]
            def_plugin_args = plugin_cfg.get('plugin_args')
            if def_plugin_args and 'qsub_args' in def_plugin_args:
                # Retain this workflow's default even if a node defined
                # in this workflow is included in a parent workflow.
                def_plugin_args['overwrite'] = True
                self.logger.debug("Workflow %s default node plug-in parameters:"
                                  " %s." % (workflow.name, def_plugin_args))
        else:
            def_plugin_args = None

        # The nodes defined in this workflow start with the workflow name.
        prefix = workflow.name + '.'
        # Configure all node inputs.
        nodes = [workflow.get_node(name)
                 for name in workflow.list_node_names()]
        for node in nodes:
            # An input {field: value} dictionary to format a debug log message.
            input_dict = {}
            # The node configuration.
            node_cfg = self._node_configuration(workflow, node)
            # Set the node inputs or plug-in argument.
            for attr, value in node_cfg.iteritems():
                if attr == 'plugin_args':
                    # If the workflow is on a cluster and the node plug-in
                    # arguments do not overwrite the default plug-in arguments,
                    # then append the node plug-in arguments to the default
                    # and set the overwrite flag. This ensures that the node
                    # plug-in arguments take precedence over the defaults and
                    # that the arguments are retained if the node is included
                    # in a parent workflow.
                    if self.is_distributable:
                        if ('qsub_args' in value and
                                'overwrite' not in value and
                                'qsub_args' in def_plugin_args):
                            qsub_args = value['qsub_args']
                            def_qsub_args = def_plugin_args['qsub_args']
                            value['qsub_args'] = def_qsub_args + ' ' + qsub_args
                            value['overwrite'] = True
                        setattr(node, attr, value)
                        self.logger.debug("%s workflow node %s plugin"
                                          " arguments: %s" %
                                          (workflow.name, node, value))
                else:
                    # The current attribute value.
                    if hasattr(node.inputs, attr):
                        current = getattr(node.inputs, attr)
                    elif hasattr(node, attr):
                        current = getattr(node, attr)
                    else:
                        raise WorkflowError("The node %s does not have an"
                                            " attribute or input field %s" %
                                            (node, attr))
                    # If the config value differs from the default
                    # value, then capture the config entry for update.
                    if value != current:
                        input_dict[attr] = value

            # If:
            # 1) the configuration specifies a default,
            # 2) the node itself is not configured with plug-in arguments, and
            # 3) the node is defined in this workflow as opposed to a child
            #    workflow (i.e., the node name prefix is this workflow name),
            # then set the node plug-in arguments to the default.
            delegatable = def_plugin_args and 'plugin_args' not in node_cfg
            if delegatable and str(node).startswith(prefix):
                node.plugin_args = def_plugin_args

            # If a field was set to a config value, then print the config
            # setting to the log.
            if input_dict:
                self._set_node_inputs(node, **input_dict)
                self.logger.debug("The following %s workflow node %s inputs"
                                   " were set from the configuration: %s" %
                                   (workflow.name, node, input_dict))

    def _set_node_inputs(self, node, **kwargs):
        """
        Sets the given node attributes. The input attributes can be
        either a node input interface field, e.g. the
        :class:`qipipe.interfaces.copy` *dest* field, or an attribute
        of the node itself, e.g. *run_without_submitting*.
        
        :param node: the target node
        :param kwargs: the input {attribute: value} dictionary
        """
        # The node interface input traits.
        traits = node.inputs.traits()
        # The input interface {attribute: value} dictionary.
        input_dict = {attr: kwargs[attr] for attr in kwargs if attr in traits}
        # The input interface attributes.
        input_attrs = set(input_dict.iterkeys())
        # The {attribute: dependencies} requirement dictionary.
        req_dict = {attr: set(traits[attr].requires).intersection(input_attrs)
                    for attr in input_dict
                    if traits[attr].requires}
        # The dependency graph.
        req_grf = nx.DiGraph()
        # The dependency graph nodes.
        req_grf.add_nodes_from(input_dict)
        # The dependency graph edges.
        for attr, reqs in req_dict.iteritems():
            for req in reqs:
                req_grf.add_edge(req, attr)
        # Sort the input interface attributes by dependency.
        sorted_attrs = nx.topological_sort(req_grf)
        # Set the node input field values.
        for attr in sorted_attrs:
            setattr(node.inputs, attr, input_dict[attr])
        # Set the node attribute values.
        node_attrs = (attr for attr in kwargs if not attr in input_dict)
        for attr in node_attrs:
            setattr(node, attr, kwargs[attr])

    def _node_configuration(self, workflow, node):
        """
        Returns the {parameter: value} dictionary defined for the given
        node in this WorkflowBase's configuration. The configuration topic
        is determined as follows:

        * the node class, as described in :meth:`_interface_configuration`

        * the node name, qualified by the node hierarchy if the node is
          defined in a child workflow

        :param workflow: the parent or nested workflow object
        :param node: the interface class to check
        :return: the corresponding {field: value} dictionary
        """
        return (self._interface_configuration(node.interface.__class__) or
         self._node_name_configuration(workflow, node) or EMPTY_DICT)

    def _node_name_configuration(self, workflow, node):
        """
        Returns the {parameter: value} dictionary defined for the given
        node name, qualified by the node hierarchy if the node is
        defined in a child workflow.

        :param workflow: the active workflow
        :param node: the interface class to check
        :return: the corresponding {field: value} dictionary
        """
        if node._hierarchy == workflow.name:
            return self.configuration.get(node.name)
        else:
            return self.configuration.get(node.fullname)

    def _interface_configuration(self, klass):
        """
        Returns the {parameter: value} dictionary defined for the given
        interface class in this WorkflowBase's configuration. The
        configuration topic matches the module path of the interface class
        name. The topic can elide the ``interfaces`` or ``interface`` prefix,
        e.g.:

            [nipype.interfaces.ants.AverageImages]
            [ants.AverageImages]

        both refer to the Nipype ANTS AverageImages wrapper interface.

        :param node: the interface class to check
        :return: the corresponding {field: value} dictionary
        """
        topic = "%s.%s" % (klass.__module__, klass.__name__)
        if topic in self.configuration:
            return self.configuration[topic]
        elif WorkflowBase.INTERFACE_PREFIX_PAT.match(topic):
            # A parent module might import the class. Therefore,
            # strip out the last module and retry.
            abbr = WorkflowBase.INTERFACE_PREFIX_PAT.sub('', topic)
            while abbr:
                if abbr in self.configuration:
                    return self.configuration[abbr]
                match = WorkflowBase.MODULE_PREFIX_PAT.match(abbr)
                if match:
                    prefix, _, _, name = match.groups()
                    abbr = prefix + name
                else:
                    abbr = None



"""
This module updates the qiprofile database scan information
from a XNAT experiment.
"""

class ScanError(Exception):
    pass


def update(session, xscan):
    """
    Updates the scan content for the given qiprofile session
    database object from the given XNAT scan object.

    :param session: the target qiprofile Session to update
    :param xscan: the XNAT scan object
    """
    pass



import os
import logging
from nipype.pipeline import engine as pe
from nipype.interfaces.utility import (IdentityInterface, Function)
from nipype.interfaces.dcmstack import DcmStack
from ..interfaces import (Gate, FixDicom, Compress, XNATFind, XNATUpload)
import qixnat
from .workflow_base import WorkflowBase
from qiutil.logging import logger
from ..staging import iterator


def set_workflow_iterables(exec_wf, scan_input, dest=None):
    """
    Sets the given execution workflow iterables.
    The execution workflow must have the same iterable
    node names and fields as the :class:`StagingWorkflow` workflow.

    :param exec_wf: the workflow to execute
    :param scan_input: the :class:`qipipe.staging.iterator.ScanInput`
        object
    :param dest: the TCIA staging destination directory (default is
        the current working directory)
    """
    # The input volumes to stage.
    volumes = scan_input.iterators.dicom.keys()
    # Make a staging area subdirectory for each volumes.
    stg_dict = _create_staging_area(scan_input.subject, scan_input.session,
                                    scan_input.scan, volumes, dest)
    # The staging destination directories are pair-wise synchronized
    # with the input volumes.
    dests = [stg_dict[volume] for volume in volumes]
    iterables = dict(volume=volumes, dest=dests).items()

    # Set the volume iterator inputs.
    iter_volume = exec_wf.get_node('iter_volume')
    iter_volume.iterables = iterables
    # Iterate over the volume and dest input fields in lock-step.
    iter_volume.synchronize = True

    # Set the DICOM file iterator inputs.
    iter_dicom = exec_wf.get_node('iter_dicom')
    iter_dicom.itersource = ('iter_volume', 'volume')
    iter_dicom.iterables = ('dicom_file', scan_input.iterators.dicom)


def _create_staging_area(subject, session, scan, volumes, dest=None):
    """
    :param subject: the subject name
    :param session: the session name
    :param scan: the scan number
    :param volumes: the input volumes
    :param dest: the TCIA staging destination directory (default is
        the current working directory)
    :return: the {volume number: directory} dictionary
    """
    # The staging location.
    dest = os.path.abspath(dest) if dest else os.getcwd()
    # Collect the {volume: destination} dictionary.
    return {volume: _make_volume_staging_directory(dest, subject, session,
                                                   scan, volume)
            for volume in volumes}


def _make_volume_staging_directory(dest, subject, session, scan, volume):
    """
    Returns the dest/subject/session/volume directory path in which to
    place DICOM files for TCIA upload. Creates the directory, if
    necessary.

    :return: the target volume directory path
    """
    path = os.path.join(dest, subject, session, 'scan', str(scan), 'volume',
                        str(volume))
    if not os.path.exists(path):
        os.makedirs(path)

    return path


class StagingWorkflow(WorkflowBase):
    """
    The StagingWorkflow class builds and executes the staging Nipype workflow.
    The staging workflow includes the following steps:

    - Group the input DICOM images into volume.

    - Fix each input DICOM file header using the
      :class:`qipipe.interfaces.fix_dicom.FixDicom` interface.

    - Compress each corrected DICOM file.

    - Upload each compressed DICOM file into XNAT.

    - Stack each new volume's 2-D DICOM files into a 3-D volume NIfTI file
      using the DcmStack_ interface.

    - Upload each new volume stack into XNAT.

    - Make the CTP_ QIN-to-TCIA subject id map.

    - Collect the id map and the compressed DICOM images into a target
      directory in collection/subject/session/volume format for TCIA
      upload.

    The staging workflow input is the *input_spec* node consisting of
    the following input fields:

    - *collection*: the collection name

    - *subject*: the subject name

    - *session*: the session name

    - *scan*: the scan number

    The staging workflow has two iterables:

    - the *iter_volume* node with input fields *volume* and *dest*

    - the *iter_dicom* node with input fields *volume* and *dicom_file*

    These iterables must be set prior to workflow execution. The
    *iter_volume* *dest* input is the destination directory for
    the *iter_volume* *volume*.

    The *iter_dicom* node *itersource* is the ``iter_volume.volume``
    field. The ``iter_dicom.dicom_file`` iterables is set to the
    {volume: [DICOM files]} dictionary.

    The DICOM files to upload to TCIA are placed in the destination
    directory in the following hierarchy:

        ``/path/to/dest/``
          *subject*\ /
            *session*\ /
              ``volume``\ *volume number*\ /
                *file*
                ...

    where:

    * *subject* is the subject name, e.g. ``Breast011``

    * *session* is the session name, e.g. ``Session03``

    * *volume number* is determined by the
      :attr:`qipipe.staging.collection.Collection.patterns`
      :attr:`qipipe.staging.collection.Patterns.volume` DICOM tag

    * *file* is the DICOM file name

    The staging workflow output is the *output_spec* node consisting
    of the following output field:

    - *image*: the 3D volume stack NIfTI image file

    Note: Concurrent XNAT upload fails unpredictably due to one of
        the causes described in the ``qixnat.facade.XNAT.find`` method
        documentation.

        The errors are addressed by the following measures:
        * setting an isolated pyxnat cache_dir for each execution node
        * serializing the XNAT find-or-create access points with JoinNodes
        * increasing the SGE submission resource parameters. The following
          setting is adequate:
             h_rt=02:00:00,mf=32M

    .. _CTP: https://wiki.cancerimagingarchive.net/display/Public/Image+Submitter+Site+User%27s+Guide
    .. _DcmStack: http://nipy.sourceforge.net/nipype/interfaces/generated/nipype.interfaces.dcmstack.html
    """

    def __init__(self, **kwargs):
        """
        If the optional configuration file is specified, then the workflow
        settings in that file override the default settings.

        :param kwargs: the :class:`qipipe.pipeline.workflow_base.WorkflowBase`
            initializer keyword arguments
        """
        super(StagingWorkflow, self).__init__(logger=logger(__name__), **kwargs)

        # Make the workflow.
        self.workflow = self._create_workflow()
        """
        The staging workflow sequence described in
        :class:`qipipe.pipeline.staging.StagingWorkflow`.
        """

    def set_inputs(self, scan_input, dest=None):
        """
        Sets the staging workflow inputs for the *input_spec* node
        and the iterables.

        :param scan_input: the :class:`qipipe.staging.iterator.ScanInput`
            object
        :param dest: the TCIA staging destination directory (default is
            the current working directory)
        """
        # Set the top-level inputs.
        input_spec = self.workflow.get_node('input_spec')
        input_spec.inputs.collection = scan_input.collection
        input_spec.inputs.subject = scan_input.subject
        input_spec.inputs.session = scan_input.session
        input_spec.inputs.scan = scan_input.scan

        # Set the iterables.
        set_workflow_iterables(self.workflow, scan_input, dest)

    def run(self):
        """Executes the staging workflow."""
        self._run_workflow(self.workflow)

    def _create_workflow(self):
        """
        Makes the staging workflow described in
        :class:`qipipe.pipeline.staging.StagingWorkflow`.
        :return: the new workflow
        """
        self.logger.debug('Creating the DICOM processing workflow...')

        # The Nipype workflow object.
        workflow = pe.Workflow(name='staging')

        # The workflow input.
        in_fields = ['collection', 'subject', 'session', 'scan']
        input_spec = pe.Node(IdentityInterface(fields=in_fields),
                             name='input_spec')
        self.logger.debug("The %s workflow input node is %s with fields %s" %
                         (workflow.name, input_spec.name, in_fields))

        # Create the scan, if necessary. The gate blocks upload until the
        # scan is created.
        find_scan_xfc = XNATFind(project=self.project, modality='MR',
                                 create=True)
        find_scan = pe.Node(find_scan_xfc, name='find_scan')
        workflow.connect(input_spec, 'subject', find_scan, 'subject')
        workflow.connect(input_spec, 'session', find_scan, 'session')
        workflow.connect(input_spec, 'scan', find_scan, 'scan')
        scan_gate_xfc = Gate(fields=['scan', 'xnat_id'])
        scan_gate = pe.Node(scan_gate_xfc, name='scan_gate')
        workflow.connect(input_spec, 'scan', scan_gate, 'scan')
        workflow.connect(find_scan, 'xnat_id', scan_gate, 'xnat_id')

        # The volume iterator.
        iter_volume_fields = ['volume', 'dest']
        iter_volume = pe.Node(IdentityInterface(fields=iter_volume_fields),
                              name='iter_volume')
        self.logger.debug("The %s workflow volume iterable node is %s with"
                           " fields %s" % (workflow.name, iter_volume.name,
                                           iter_volume_fields))

        # The DICOM file iterator.
        iter_dicom_fields = ['volume', 'dicom_file']
        iter_dicom = pe.Node(IdentityInterface(fields=iter_dicom_fields),
                             name='iter_dicom')
        self.logger.debug("The %s workflow DICOM iterable node is %s with"
                           " iterable source %s and iterables"
                           " ('%s', {%s: [%s]})" %
                           (workflow.name, iter_dicom.name, iter_volume.name,
                           'dicom_file', 'volume', 'DICOM files'))
        workflow.connect(iter_volume, 'volume', iter_dicom, 'volume')

        # Fix the DICOM tags.
        fix_dicom = pe.Node(FixDicom(), name='fix_dicom')
        workflow.connect(input_spec, 'collection', fix_dicom, 'collection')
        workflow.connect(input_spec, 'subject', fix_dicom, 'subject')
        workflow.connect(iter_dicom, 'dicom_file', fix_dicom, 'in_file')

        # Compress the corrected DICOM files.
        compress_dicom = pe.Node(Compress(), name='compress_dicom')
        workflow.connect(fix_dicom, 'out_file', compress_dicom, 'in_file')
        workflow.connect(iter_volume, 'dest', compress_dicom, 'dest')

        # Collect the compressed session DICOM files as follows:
        # * The volume DICOM files are collected into a list.
        # * The volume DICOM file lists are merged into a scan
        #   DICOM file list.
        # * The combined scan DICOM file list is uploaded to XNAT.
        #
        # The collection involves a two-step JoinNode, first on
        # the volume, then on the scan. The second scan JoinNode
        # calls a merge function to merge the lists from all of
        # the first JoinNodes.
        #
        # The session files must be uploaded in a single task to
        # work around the following XNAT error:
        #
        # * Concurrent upload to a given XNAT resource corrupts the
        #   XNAT file object. The files are copied in the archive
        #   location, but the file objects are not created.
        #
        # Since concurrent XNAT upload is not supported, all of the
        # compressed DICOM files must be collected into a single
        # list which is uploaded in a single upload task when
        # deployed in a clustered environment such as SGE.
        collect_vol_dicom_xfc = IdentityInterface(fields=['dicom_files'])
        collect_vol_dicom = pe.JoinNode(collect_vol_dicom_xfc,
                                        joinsource='iter_dicom',
                                        joinfield='dicom_files',
                                        name='collect_vol_dicom')
        workflow.connect(compress_dicom, 'out_file',
                         collect_vol_dicom, 'dicom_files')
        collect_scan_dicom_xfc = Function(input_names=['lists'],
                                          output_names=['out_list'],
                                          function=merge)
        collect_scan_dicom = pe.JoinNode(collect_scan_dicom_xfc,
                                         joinsource='iter_volume',
                                         joinfield='lists',
                                         name='collect_scan_dicom')
        workflow.connect(collect_vol_dicom, 'dicom_files',
                         collect_scan_dicom, 'lists')

        # Upload the compressed scan DICOM files.
        upload_dicom_xfc = XNATUpload(project=self.project, resource='DICOM',
                                      skip_existing=True, modality='MR')
        upload_dicom = pe.Node(upload_dicom_xfc, name='upload_dicom')
        workflow.connect(input_spec, 'subject', upload_dicom, 'subject')
        workflow.connect(input_spec, 'session', upload_dicom, 'session')
        workflow.connect(scan_gate, 'scan', upload_dicom, 'scan')
        workflow.connect(collect_scan_dicom, 'out_list',
                         upload_dicom, 'in_files')

        # The volume file name format.
        vol_fmt_xfc = Function(input_names=['collection'],
                               output_names=['format'],
                               function=volume_format)
        vol_fmt = pe.Node(vol_fmt_xfc, name='volume_format')
        workflow.connect(input_spec, 'collection', vol_fmt, 'collection')

        # Stack the scan volume into a 3D NIfTI file.
        stack_xfc = DcmStack(embed_meta=True)
        stack = pe.JoinNode(stack_xfc, joinsource='iter_dicom',
                            joinfield='dicom_files', name='stack')
        workflow.connect(fix_dicom, 'out_file', stack, 'dicom_files')
        workflow.connect(vol_fmt, 'format', stack, 'out_format')

        # Force the T1 3D upload to follow DICOM upload to avoid the
        # concurrency conflict described in :class:`StagingWorkflow`.
        # This gate task serializes upload to prevent potential XNAT access
        # conflicts.
        upload_3d_gate_xfc = Gate(fields=['out_file', 'xnat_files'])
        upload_3d_gate = pe.Node(upload_3d_gate_xfc, name='upload_3d_gate')
        workflow.connect(upload_dicom, 'xnat_files', upload_3d_gate, 'xnat_files')
        workflow.connect(stack, 'out_file', upload_3d_gate, 'out_file')

        # Upload the 3D NIfTI stack files to XNAT.
        upload_3d_xfc = XNATUpload(project=self.project, resource='NIFTI',
                                   skip_existing=True, modality='MR')
        upload_3d = pe.JoinNode(upload_3d_xfc, joinsource='iter_volume',
                                joinfield='in_files', name='upload_3d')
        workflow.connect(input_spec, 'subject', upload_3d, 'subject')
        workflow.connect(input_spec, 'session', upload_3d, 'session')
        workflow.connect(scan_gate, 'scan', upload_3d, 'scan')
        # 3D upload is gated by DICOM upload.
        workflow.connect(upload_3d_gate, 'out_file', upload_3d, 'in_files')

        # The output is the 3D NIfTI stack file. Make an intermediate Gate
        # node to ensure that upload is completed before setting the output
        # field.
        output_gate_xfc = Gate(fields=['image', 'xnat_files'])
        output_gate = pe.Node(output_gate_xfc, name='output_gate')
        workflow.connect(stack, 'out_file', output_gate, 'image')
        workflow.connect(upload_3d, 'xnat_files', output_gate, 'xnat_files')

        # Make the output a Gate node to work around the following Nipype
        # bug:
        # * Nipype overzealously prunes an IdentityInterface node as
        #   extraneous, even if it is connected in a parent workflow.
        # TODO - verify that this is still the case
        output_spec_xfc = Gate(fields=['image'])
        output_spec = pe.Node(output_spec_xfc, name='output_spec')
        workflow.connect(output_gate, 'image', output_spec, 'image')

        # Instrument the nodes for cluster submission, if necessary.
        self._configure_nodes(workflow)

        self.logger.debug("Created the %s workflow." % workflow.name)
        # If debug is set, then diagram the workflow graph.
        if self.logger.level <= logging.DEBUG:
            self.depict_workflow(workflow)

        return workflow


def volume_format(collection):
    """
    :param collection: the collection name
    :return: the volume file name format
    """
    from qipipe.staging import collections

    coll = collections.with_name(collection)

    return "volume%%(%s)03d" % coll.patterns.volume


def merge(lists):
    """
    Merges the given lists. This function works around the following
    Nipype Merge node limitation:

    * The Nipype Merge initializer requires the number of lists to merge.

    This merge function accepts an arbitrary number of lists.

    :param lists: this lists to merge
    :return: the merged lists
    """
    merger = lambda x,y: x + y

    return reduce(merger, lists)



from nipype.interfaces.base import (
    traits, BaseInterfaceInputSpec, TraitedSpec)
from nipype.interfaces.io import IOBase


class LookupInputSpec(BaseInterfaceInputSpec):
    key = traits.Any(mandatory=True, desc='The dictionary key')

    dictionary = traits.Dict(mandatory=True, desc='The dictionary')


class LookupOutputSpec(TraitedSpec):
    value = traits.Any(desc='The lookup value')


class Lookup(IOBase):
    """
    The Lookup Interface wraps a dictionary look-up.

    Example:

    >>> from qipipe.interfaces import Lookup
    >>> lookup = Lookup(key='a', dictionary=dict(a=1, b=2))
    >>> result = lookup.run()
    >>> result.outputs.value
    1
    """

    input_spec = LookupInputSpec

    output_spec = LookupOutputSpec

    def _run_interface(self, runtime):
        # The output name => value dictionary result.
        self._result = self.inputs.dictionary[self.inputs.key]

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['value'] = self._result

        return outputs



import re
from .staging_error import StagingError
from . import collections


class ROIPatterns(object):
    """The ROI file name matching patterns."""

    def __init__(self, glob, regex):
        self.glob = glob
        """The ROI file glob pattern."""
        
        self.regex = regex
        """The ROI file name match regular expression."""

    def __repr__(self):
        return str(dict(glob=self.glob, regex=self.regex))


class ScanPatterns(object):
    """The scan file name matching patterns."""

    def __init__(self, dicom, roi=None):
        self.dicom = dicom
        """The DICOM file match *glob* and *regex* patterns."""
        
        self.roi = roi
        """The :class:`ROIPatterns` object."""

    def __repr__(self):
        return str(dict(dicom=self.dicom, roi=self.roi))


class Patterns(object):
    """The collection file name and DICOM tag patterns."""
    
    def __init__(self, subject, session, scan, volume):
        """
        :param subject: the subject directory name match regular expression
        :param session: the session directory name match regular expression
        :param scan the {scan number: :class:`ScanPatterns} dictionary
        :option volume: the DICOM tag which identifies a scan volume
        """
        self.subject = subject
        """The subject directory match pattern."""

        self.session = session
        """The subject directory match pattern."""

        self.scan = scan
        """The {scan number: :class:`ScanPatterns`} dictionary."""

        self.volume = volume
        """The DICOM tag which identifies a scan volume."""


class Collection(object):
    """The image collection."""

    def __init__(self, name, **opts):
        """
        :param name: `self.name`
        :param opts: the :class:`Patterns` attributes
        """
        self.name = name
        """The collection name."""

        self.patterns = Patterns(**opts)
        """The file and DICOM meta-data patterns."""

        # Add this collection to the collections extent.
        collections.add(self)


import os
from six.moves.configparser import ConfigParser as Config

CFG_FILE = os.path.join(
    os.path.dirname(__file__), '..', 'conf', 'ctp.cfg')


def ctp_collection_for_name(name):
    """
    :param name: the QIN collection name
    :return: the CTP collection name
    """
    return ctp_config().get('CTP', name)


def ctp_config():
    if not hasattr(ctp_config, 'instance'):
        ctp_config.instance = Config()
        ctp_config.instance.read(CFG_FILE)
    return ctp_config.instance



"""
OHSU - This module wraps the proprietary OHSU AIRC ``bolero_mask_conv``
utility. ``bolero_mask_conv`` converts a proprietary OHSU format
mask file into a NIfTI mask file.
"""
import os
from os import path
from glob import glob
import traits.api as traits
from nipype.interfaces.base import (TraitedSpec, CommandLine,
                                    CommandLineInputSpec)
from nipype.interfaces.traits_extension import Undefined


class ConvertBoleroMaskInputSpec(CommandLineInputSpec):
    time_series = traits.File(desc='Input 4D DCE series NIfTI file',
                              mandatory=True, exists=True, position=1,
                              argstr='%s')
    slice_sequence_number = traits.Int(
        desc='One-based sequential slice number', mandatory=True, position=2,
        argstr='%d'
    )
    in_file = traits.Str(desc='BOLERO .bqf mask file', mandatory=True,
                         position=3, argstr='\"%s\"')
    out_base = traits.Str(desc='Output file base name without extension',
                          argstr='-o %s')


class ConvertBoleroMaskOutputSpec(TraitedSpec):
    out_file = traits.File(desc='NIfTI mask file', exists=True)


class ConvertBoleroMask(CommandLine):
    """
    Interface to the proprietary OHSU AIRC ``bolero_mask_conv`` utility.
    """

    _cmd = 'bolero_mask_conv'
    input_spec = ConvertBoleroMaskInputSpec
    output_spec = ConvertBoleroMaskOutputSpec

    def __init__(self, **inputs):
        super(ConvertBoleroMask, self).__init__(**inputs)

    def _list_outputs(self):
        outputs = self._outputs().get()

        # The default output base name is slice_<slice>_lesion_mask.
        out_base = (self.inputs.out_base or
                    "slice_%d_lesion_mask" % self.inputs.slice_sequence_number)
        # The output is compressed.
        out_file = out_base + '.nii.gz'
        # Expand the output path.
        outputs['out_file'] = os.path.abspath(out_file)

        return outputs



#! /usr/global/bin/python

import sys, argparse, math
from multiprocessing import Pool
import numpy as np
import nibabel as nb
from dcmstack.dcmmeta import NiftiWrapper

class const_array(object):
    def __init__(self, index_result, length=None):
        self.index_result = index_result
        self.length = length
        
    @property
    def flat(self):
        return self.__iter__()
        
    def __iter__(self):
        count = 0
        while True:
            if not self.length is None and count == self.length:
                break
            yield self.index_result
            count += 1
        
    def __getitem__(self, index):
        return self.index_result

def dce_to_r1(dce_nw, r1_0, dce_fa=None, dce_tr=None, baseline_end=1, 
              mask=None, pool=None):
    '''Convert a DCE time series to a series of R1 maps.
    
    Parameters
    ----------
    dce_nw : NiftiWrapper
        The DCE time series data
        
    r1_0 : Array or float
        The initial pre-contrast R1 value(s). Can be array with same 
        spatial shape as the DCE data, or a constant value.
        
    dce_fa : float
        The flip angle (in degrees) used to acquire the DCE data. If 
        not specified will try to get from 'dce_nw' meta data.
        
    dce_tr : float
        The repetition time (in milliseconds) used to acquire the DCE 
        data. If not specified will try to get from 'dce_nw' meta data.
        
    baseline_end : int
        The end index for the pre-contrast data in 'dce_nw'.
        
    mask : array
        Optional mask to restrict the voxels that are processed.
        
    pool : multiprocessing.Pool
        If provided, will use to do the computation in parallel
        
    Returns
    -------
    r1_series : array
        A 4D series of R1 maps.
    
    '''
    #Lookup any parameters not provided
    if dce_fa is None:
        dce_fa = dce_nw.get_meta('FlipAngle')
    if dce_tr is None:
        dce_tr = dce_nw.get_meta('RepetitionTime')
    if dce_fa is None or dce_tr is None:
        raise ValueError("Flip angle and TR image parameters not "
                         "specified or found in embedded meta data")
                         
    #Convert units
    dce_fa = math.radians(dce_fa)
    dce_tr /= 1000.
    
    #Load DCE array
    dce_data = dce_nw.nii_img.get_data()
    shape = dce_data.shape
    if len(shape) != 4:
        raise ValueError("The DCE data must be 4D")
    vol_shape = shape[:3]
    n_time = shape[3]
    
    #Handle the r1_0 input
    if isinstance(r1_0, float):
        r1_0 = const_array(r1_0)
    else:
        if r1_0.shape != vol_shape:
            raise ValueError("r1_0 must have same spatial shape as DCE data")
        r1_0 = r1_0.flat
    
    #Handle the mask data
    if mask is None:
        mask = np.ones(vol_shape, dtype=np.int8)
    elif mask.shape != vol_shape:
        raise ValueError("Mask must have same spatial shape as DCE data")
        
    #Allocate result array
    r1_series = np.zeros(shape, dtype=np.float32)
    
    #Precompute a couple of things
    sin_fa = math.sin(dce_fa)
    cos_fa = math.cos(dce_fa)
    
    #Iterate through voxels
    dce_flat = dce_data.flat
    r1_flat = r1_series.flat
    for flat_idx, mask_val in enumerate(mask.flat):
        if mask_val == 0:
            continue
        
        #Pull out values/slices we need
        start_idx = flat_idx * n_time
        end_idx = start_idx + n_time
        dce_signal = dce_flat[start_idx:end_idx]
        baseline_avg = np.mean(dce_signal[:baseline_end])
        r1_0_val = r1_0[flat_idx]
        
        #Calculate M0
        m_0 = ((baseline_avg * 
                (1. - math.exp(-dce_tr * r1_0_val) * cos_fa)) / 
               ((1. - math.exp(-dce_tr * r1_0_val)) * sin_fa)
              )
              
        if m_0 <= 0.0:
            continue
              
        #Calculate R1 values
        for time_idx, signal_val in enumerate(dce_signal):
            sig_to_m0_ratio = signal_val / m_0
            numerator = sig_to_m0_ratio - sin_fa
            denominator = (sig_to_m0_ratio * cos_fa) - sin_fa
            #If the logarithm is going to be positive or undefined, 
            #just set to zero
            if (numerator * denominator < 0 or 
               numerator / denominator > 1.0):
                continue
            r1_flat[start_idx + time_idx] = \
                (-math.log(numerator / denominator) / dce_tr)
    
    return r1_series
    
prog_descrip = """Convert DCE time series of signal intensities to 
series of R1 values.
"""

prog_epilog = """If there is embedded meta data, any unspecified flip 
angle or repetition time parameters will be looked up there."""

def main(argv=sys.argv):
    arg_parser = argparse.ArgumentParser(description=prog_descrip, 
                                         epilog=prog_epilog)
    arg_parser.add_argument('dce_nii', nargs=1, 
                            help=("4D DCE series"))
    arg_parser.add_argument('r1_0', nargs=1, 
                            help=("Constant or 3D map of R1 value(s)"))
    arg_parser.add_argument('-b', '--base-end', default=1, type=int,
                            help="End index for baseline signal")
    arg_parser.add_argument('-m', '--mask', help="Mask image")
    arg_parser.add_argument('--dce-fa', type=float, help=("Flip angle "
                            "for DCE acquisition"))
    arg_parser.add_argument('--dce-tr', type=float, 
                            help=("Repetition time for DCE acquision"))
    args = arg_parser.parse_args(argv[1:])
    
    #Load the DCE data
    dce_nw = NiftiWrapper(nb.load(args.dce_nii[0]), make_empty=True)
    
    #Setup mask
    if args.mask:
        mask = nb.load(args.mask).get_data()
    else:
        mask = None
        
    #Handle the r1_0 input
    try:
        r1_0 = float(args.r1_0[0])
    except:
        r1_0 = nb.load(args.r1_0[0]).get_data()
        
    #Compute R1 series 
    r1_series = dce_to_r1(dce_nw, r1_0, args.dce_fa, args.dce_tr, args.base_end, mask)
    
    #Write out result
    out_nii = nb.Nifti1Image(r1_series, dce_nw.nii_img.get_affine())
    nb.save(out_nii, 'r1_series.nii.gz')
    
if __name__ == "__main__":
    sys.exit(main()) 




from nipype.interfaces.base import (traits, BaseInterfaceInputSpec,
                                    TraitedSpec, BaseInterface)
from nipype.interfaces.traits_extension import isdefined
import qixnat


class XNATFindInputSpec(BaseInterfaceInputSpec):
    project = traits.Str(mandatory=True, desc='The XNAT project id')

    subject = traits.Str(mandatory=True, desc='The XNAT subject name')

    session = traits.Str(desc='The XNAT session name')

    scan = traits.Either(traits.Int, traits.Str, desc='The XNAT scan name')

    reconstruction = traits.Str(desc='The XNAT reconstruction name')

    assessor = traits.Str(desc='The XNAT assessor name')

    resource = traits.Str(desc='The XNAT resource name')

    file = traits.Str(desc='The XNAT file name')

    inout = traits.Str(desc='The XNAT resource in/out designator')

    create = traits.Bool(default=False,
                         desc='Flag indicating whether to create the XNAT'
                              ' object if it does not yet exist')

    modality = traits.Str(desc="The XNAT scan modality, e.g. 'MR'")


class XNATFindOutputSpec(TraitedSpec):
    xnat_id = traits.Str(desc='The XNAT object id')


class XNATFind(BaseInterface):
    """
    The ``XNATFind`` Nipype interface wraps the ``qixnat.facade.XNAT``
    ``find_one`` and ``find_or_create`` methods.

    :Note: concurrent XNAT operations can fail. See the
        :class:`qipipe.pipeline.staging.StagingWorkflow` note.
    """

    input_spec = XNATFindInputSpec

    output_spec = XNATFindOutputSpec

    def __init__(self, **inputs):
        super(XNATFind, self).__init__(**inputs)

    def _run_interface(self, runtime):
        # The find options.
        opts = {}
        if self.inputs.modality:
            opts['modality'] = self.inputs.modality

        # The session is optional.
        if isdefined(self.inputs.session):
            session = self.inputs.session
        else:
            session = None

        # The resource parent.
        if self.inputs.scan:
            opts['scan'] = self.inputs.scan
        elif self.inputs.reconstruction:
            opts['reconstruction'] = self.inputs.reconstruction
        elif self.inputs.assessor:
            opts['assessor'] = self.inputs.assessor

        # The resource.
        if isdefined(self.inputs.resource):
            opts['resource'] = self.inputs.resource

        # The file.
        if isdefined(self.inputs.file):
            opts['file'] = self.inputs.file

        # Delegate to the XNAT helper.
        create = isdefined(self.inputs.create) and self.inputs.create
        with qixnat.connect() as xnat:
            if create:
                obj = xnat.find_or_create(self.inputs.project,
                                          self.inputs.subject, session,
                                          **opts)
            else:
                obj = xnat.find_one(self.inputs.project, self.inputs.subject,
                                    session, **opts)
            self._xnat_id = obj.id() if obj else None

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        if self._xnat_id:
            outputs['xnat_id'] = self._xnat_id

        return outputs



"""
This module updates the qiprofile database modeling information
from a XNAT experiment.
"""

class ModelingError(Exception):
    pass


def update(session, resource):
    """
    Updates the modeling content for the given qiprofile session
    database object from the given XNAT modeling resource object.

    :param session: the target qiprofile Session to update
    :param resource: the XNAT modeling resource object
    """
    # TODO
    pass



import os
from os import path
import traits.api as traits
from nipype.interfaces.base import (TraitedSpec, CommandLine,
                                    CommandLineInputSpec)
from nipype.interfaces.traits_extension import Undefined


class MriVolClusterInputSpec(CommandLineInputSpec):
    in_file = traits.File(desc='Input file',
                          mandatory=True,
                          exists=True,
                          argstr='--in %s')
    min_thresh = traits.Float(0.0,
                              usedefault=True,
                              desc="Minimum threshold value",
                              argstr='--thmin %s')
    max_thresh = traits.Float(desc="Maximum threshold value",
                              argstr='--thmax %s')
    min_size = traits.Float(desc="Minimum cluster size in mm^3",
                            argstr='--minsize %s')
    min_voxels = traits.Int(desc="Minimum number of voxels in cluster",
                            argstr='--minsizevox %s')
    out_clusters_name = traits.Str('clusters.nii.gz',
                                   usedefault=True,
                                   desc="Name out output image with cluster labels",
                                   argstr="--ocn %s")
    out_masked_name = traits.Str('masked.nii.gz',
                                 usedefault=True,
                                 desc="Name of output masked image",
                                 argstr="--out %s")


class MriVolClusterOutputSpec(TraitedSpec):
    out_cluster_file = traits.File(
        desc="Output image containing cluster labels")
    out_masked_file = traits.File(desc="Input image masked by clusters")


class MriVolCluster(CommandLine):
    """
    MriVolCluster encapsulates the FSL mri_volcluster_ command.
    
    .. _mri_volcluster: http://ftp.nmr.mgh.harvard.edu/fswiki/mri_volcluster
    """

    input_spec = MriVolClusterInputSpec
    output_spec = MriVolClusterOutputSpec
    cmd = 'mri_volcluster'

    def _list_outputs(self):
        cwd = os.getcwd()
        outputs = self.output_spec().get()
        if not self.inputs.out_clusters_name is Undefined:
            outputs['out_cluster_file'] = path.join(cwd,
                                                    self.inputs.out_clusters_name)
        if not self.inputs.out_masked_name is Undefined:
            outputs['out_masked_file'] = path.join(cwd,
                                                   self.inputs.out_masked_name)
        return outputs



import os
from nipype.interfaces.base import (BaseInterfaceInputSpec, TraitedSpec,
                                    BaseInterface, File)


class TouchInputSpec(BaseInterfaceInputSpec):
    in_file = File(mandatory=True, desc='The file to touch')


class TouchOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='The touched file')


class Touch(BaseInterface):
    """
    The Touch interface emulates the Unix ``touch`` command.
    This interface is useful for stubbing out processing
    nodes during workflow development.
    """
    input_spec = TouchInputSpec

    output_spec = TouchOutputSpec

    def _run_interface(self, runtime):
        self._file = os.path.abspath(self.inputs.in_file)
        parent, _ = os.path.split(self._file)
        if parent:
            if not os.path.exists(parent):
                os.makedirs(parent)
            else:
                parent = os.getcwd()
        if os.path.exists(self._file):
            os.utime(self._file, None)
        else:
            open(self._file, 'w').close()

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._file

        return outputs



"""Command qipipe command options."""

import qiutil
from qiutil.command import add_options


def configure_log(**opts):
    # Configure the logger for this qixnat module and the qiutil module.
    qiutil.command.configure_log('qipipe', 'qixnat', 'qidicom', 'qiutil', **opts)



"""
This module updates the qiprofile database Subject pathology information
from the pathology Excel workbook file.
"""
from qiprofile_rest_client.model.clinical import (
    Surgery, BreastSurgery, BreastPathology, ModifiedBloomRichardsonGrade,
    ResidualCancerBurden, HormoneReceptorStatus, BreastGeneticExpression
)
from . import parse
from .pathology import (PathologyError, PathologyWorksheet, PathologyUpdate)

HORMONES = ['estrogen', 'progesterone']
"""The receptor status hormones."""



def read(workbook, **condition):
    """
    This is a convenience method that wraps :class:`BreastPathologyWorksheet`
    :meth:`qipipe.qiprofile.xls.Worksheet.read`.

    :param workbook: the read-only ``openpyxl`` workbook object
    :param condition: the :meth:`qipipe.qiprofile.xls.Worksheet.read`
        filter condition
    :return: the :meth:`qipipe.qiprofile.xls.Worksheet.read` rows
    """
    reader = PathologyWorksheet(workbook, BreastSurgery, BreastPathology,
                                ModifiedBloomRichardsonGrade,
                                ResidualCancerBurden,
                                BreastGeneticExpression,
                                parsers=_receptor_parsers())

    return reader.read(**condition)


def update(subject, rows):
    """
    Updates the given subject data object from the Breast pathology XLS
    rows.

    :param subject: the ``Subject`` Mongo Engine database object
        to update
    :param rows: the input pathology :meth:`read` rows list 
    """
    updater = BreastPathologyUpdate(subject)
    updater.update(rows)


class BreastPathologyUpdate(PathologyUpdate):
    """The Breast pathology update facade."""

    def __init__(self, subject):
        """
        :param subject: the ``Subject`` Mongo Engine database object
            to update
        """
        super(BreastPathologyUpdate, self).__init__(
            subject, tumor_type='Breast', pathology_class=BreastPathology,
            grade_class=ModifiedBloomRichardsonGrade
        )

    def encounter_type(self, row):
        """
        Overrides :meth:`qipipe.qiprofile.Pathology.encounter_type` to
        specialize the *intervention_type* to ``BreastSurgery``.

        :param row: the input row
        :return: the REST data model Encounter subclass
        """
        base_type = super(BreastPathologyUpdate, self).encounter_type(row)

        return BreastSurgery if base_type == Surgery else base_type

    def pathology_content(self, row):
        """
        Collects the pathology object from the given input row.
        This subclass implementation adds the non-empty embedded
        fields specific to this tumor type.
        
        :param row: the input row
        :return: the {attribute: value} content dictionary
        """
        # Delegate to PathologyUpdate for the common pathology fields.
        content = super(BreastPathologyUpdate, self).pathology_content(row)
        # The receptor status objects.
        rcptrs = _collect_hormone_receptors(row)
        if rcptrs:
            content['hormone_receptors'] = rcptrs
        # The RCB {attribute: value} content.
        rcb_content = _rcb_content(row)
        if rcb_content:
            rcb = ResidualCancerBurden(**rcb_content)
            content['rcb'] = rcb
        # The genetic expression {attribute: value} content.
        ge_content = _gene_expression_content(row)
        if ge_content:
            ge = BreastGeneticExpression(**ge_content)
            content['genetic_expression'] = ge
        
        return content


def _collect_hormone_receptors(row):
    """
    :param row: the input row
    :return: the non-empty receptor statuses
    """
    rcptrs = []
    for hormone in HORMONES:
        content = _hormone_receptor_content(hormone, row)
        if content:
            rcptr_status = HormoneReceptorStatus(**content)
            rcptr_status.append(rcptr)
    
    return rcptrs


def _hormone_receptor_content(hormone, row):
    """
    :param hormone: the target hormone
    :param row: the input row
    :return: the {attribute: value} dictionary for non-None row values
    """
    # The result {attribute: value} dictionary.
    content = {}
    # The fields to collect from the row, excluding hormone which
    # we already have.
    fields = (field for field in HormoneReceptorStatus._fields
              if field in row)
    # Collect the values.
    for field in fields:
        row_attr = '_'.join([hormone, field])
        val = row[row_attr]
        if val != None:
            content[field] = val

    return content


def _rcb_content(row):
    """
    :param row: the input row
    :return: the {attribute: value} dictionary for non-None row values
    """
    return {attr: row[attr] for attr in ResidualCancerBurden._fields
            if attr in row and row[attr] != None}


def _gene_expression_content(row):
    """
    :param row: the input row
    :return: the {attribute: value} dictionary for non-None row values
    """
    return {attr: row[attr] for attr in BreastGeneticExpression._fields
            if attr in row and row[attr] != None}


def _receptor_parsers():
    """
    The hormone receptor input fields are prefixed by the hormone name,
    e.g. ``estrogen_positive`` is the ``HormoneReceptorStatus.positive``
    field for the ``HormoneReceptorStatus`` with name ``estrogen``.
    The parsers for the hormone receptor input fields are thus given by
    the respective default ``HormoneReceptorStatus`` parsers.

    :return: the parser {attribute: function} dictionary
    """
    parsers = {}
    def_receptor_parsers = parse.default_parsers(HormoneReceptorStatus)
    for hormone in HORMONES:
        for field, parser in def_receptor_parsers.iteritems():
            attr = '_'.join([hormone, field])
            parsers[attr] = parser  

    return parsers



"""
This module updates the qiprofile database Subject pathology information
from the pathology Excel workbook file.
"""
import bunch
from qiprofile_rest_client.model.common import (Encounter, TumorExtent)
from qiprofile_rest_client.model.clinical import (
    Biopsy, Surgery, PathologyReport, TNM, TumorLocation
)
from . import parse
from .xls import Worksheet

ENCOUNTER_TYPES = {klass.__name__: klass for klass in (Biopsy, Surgery)}
"""The encounter {name: class} dictionary."""

COL_ATTRS = {
    'Patient Weight (kg)': 'weight',
    'Tumor Size Score': 'size',
    'Tumor Length (mm)': 'length',
    'Tumor Width (mm)': 'width',
    'Tumor Depth (mm)': 'depth'
}
"""
The following non-standard column-attribute associations:
* ``Patient Weight (kg)``: ``Encounter.weight`` attribute
* ``Tumor Size Score``: ``TNM.size`` attribute
* ``Tumor Length (mm)``: ``TumorExtent.length`` attribute
* ``Tumor Width (mm)``: ``TumorExtent.width`` attribute
* ``Tumor Depth (mm)``: ``TumorExtent.depth`` attribute
"""

PARSERS = dict(
    subject_number=int,
    lesion_number=int,
    body_part=lambda value: value.lower().capitalize(),
    # Wrap the functions below with a lambda as a convenience to allow
    # a forward reference to the parse functions defined later.
    intervention_type=lambda value: _parse_intervention_type(value),
    size=lambda value: _parse_tumor_size(value)
)
"""
The following parser associations:
* subject_number is an int
* intervention_type converts the string to an Encounter subclass
* body_part is capitalized
* size is a :class:`qiprofile_rest_client.clinical.TNM.Size` object
"""

SHEET = 'Pathology'
"""The worksheet name."""


class PathologyError(Exception):
    pass


def _parse_intervention_type(value):
    """
    :param value: the input string
    :return: the encounter class
    """
    value = value.capitalize()
    klass = ENCOUNTER_TYPES.get(value, None)
    if not klass:
        raise PathologyError("The pathology row intervention type is not"
                             " recognized: %s" % value)
    
    return klass


def _parse_tumor_size(value):
    return TNM.Size.parse(str(value))


class PathologyWorksheet(Worksheet):
    """The Pathology worksheet facade."""
    
    def __init__(self, workbook, *classes, **opts):
        """
        :param workbook: the :class:`qipipe.qiprofile.xls.Workbook` object
        :param classes: the subclass-specific REST data model subclasses
        :param opts: the following options:
        :option parsers: the non-standard parsers {attribute: function}
            dictionary
        :option column_attributes: the non-standard {column name: attribute}
            dictionary
        """
        # The special parsers.
        parsers = PARSERS.copy()
        # Add the subclass special parsers.
        parsers_opt = opts.get('parsers', None)
        if parsers_opt:
            parsers.update(parsers_opt)
        # The special column-attribute associations.
        col_attrs = COL_ATTRS.copy()
        # Add the subclass special associations.
        col_attrs_opt = opts.get('column_attributes', None)
        if col_attrs_opt:
            col_attrs.update(col_attrs_opt)
        # Initialize the worksheet.
        super(PathologyWorksheet, self).__init__(
            workbook, SHEET, Encounter, TumorLocation, TumorExtent, TNM,
            *classes, parsers=parsers, column_attributes=col_attrs
        )


class PathologyUpdate(object):
    """The pathology update abstract class."""

    def __init__(self, subject, tumor_type, grade_class, pathology_class):
        """
        :param subject: the ``Subject`` Mongo Engine database object
            to update
        :param tumor_type: the subclass tumor type
        :option pathology_class: the REST data model TumorPathology
            subclass
        :option grade_class: the REST data model Grade subclass
        """
        self._subject = subject
        self._tumor_type = tumor_type
        data_model_dict = dict(grade=grade_class, pathology=pathology_class)
        self._data_models = bunch.bunchify(data_model_dict)

    def update(self, rows):
        """
        Updates the subject data object from the given pathology XLS
        rows.

        :param rows: the input pathology :meth:`read` rows list 
        """
        # Update the encounter rows.
        for enc_rows in self._encounter_rows(rows):
            self._update_encounter(enc_rows)

    def _encounter_rows(self, rows):
        """
        Partition the rows by encounter.
        
        :param rows: the input rows
        :yield: each encounter's lesion rows 
        """
        enc_rows = []
        for row in rows:
            if row.lesion_number == 1 and enc_rows:
                yield enc_rows
                enc_rows = []
            enc_rows.append(row)
        if enc_rows:
            yield enc_rows

    def _update_encounter(self, rows):
        """
        :param rows: the input pathology :meth:`read` rows for the
            encounter
        """
        # The lesion 1 row has the encounter information.
        master_row = rows[0]
        # There must be an intervention type.
        if not master_row.intervention_type:
            raise PathologyError("The pathology input row is missing the"
                                 "intervention type")
        if not master_row.date:
            raise PathologyError("The pathology input row is missing the"
                                 "date")
        # The encounter object for the input encounter type.
        enc_type = self.encounter_type(master_row)
        enc = self._encounter_for(enc_type, master_row.date)
        # Update the encounter.
        self.update_encounter(enc, rows)

    def update_encounter(self, encounter, rows):
        """
        Update the encounter object from the given input row.
        This base implementation sets the encounter attribute values
        from the matching input row attribute value and calls
        :meth:`update_pathology` to update the pathology.
        Other updates are a subclass responsibility.
        
        :param encounter: the encounter object
        :param rows: the input pathology :meth:`read` rows for the
            encounter
        """
        # The lesion 1 row has the encounter information.
        master_row = rows[0]
        for attr in encounter.__class__._fields:
            if attr in master_row:
                setattr(encounter, attr, master_row[attr])
        # Delegate to the possibly overridden pathology_content
        # method to collect the pathology content.
        path_content = (self.pathology_content(row) for row in rows)
        tumors = [self._data_models.pathology(**content)
                  for content in path_content if path_content]
        if tumors:
            encounter.pathology = PathologyReport(tumors=tumors)

    def pathology_content(self, row):
        """
        Collects the TumorPathology content from the given input row.
        This base implementation collects the pathology attribute
        values from the matching input row attribute value. Other
        updates are a subclass responsibility.
        
        :param row: the input row
        :return: the {attribute: value} content dictionary
        """
        # The TumorPathology subclass.
        path_cls = self._data_models.pathology
        # The non-embedded content.
        path_content = {attr: row[attr] for attr in path_cls._fields
                        if attr in row and row[attr] != None}
        # The TNM non-embedded content.
        tnm_content = {attr: row[attr] for attr in TNM._fields
                       if attr in row and row[attr] != None}
        # THe TNM grade.
        grade_cls = self._data_models.grade
        grade_content = {attr: row[attr] for attr in grade_cls._fields
                         if attr in row and row[attr] != None}
        if grade_content:
            tnm_content['grade'] = grade_cls(**grade_content)
        # The TNM.
        if tnm_content:
            path_content['tnm'] = TNM(tumor_type=self._tumor_type,
                                      **tnm_content)
        # The tumor location.
        location_content = {attr: row[attr] for attr in TumorLocation._fields
                            if row.get(attr) != None}
        if location_content:
            path_content['location'] = TumorLocation(**location_content)
        # The tumor extent. The row.get condition ignores both zero
        # and None values.
        extent_content = {attr: row[attr] for attr in TumorExtent._fields
                          if row.get(attr)}
        if extent_content:
            path_content['extent'] = TumorExtent(**extent_content)

        return path_content

    def _encounter_for(self, klass, date):
        """
        :param klass: the encounter class
        :param date: the encounter date
        :return: the existing or new encounter object
        """
        # The encounter list
        encs = self._subject.encounters
        # Find the matching encounter, if it exists.
        enc_iter = (enc for enc in encs
                    if isinstance(enc, klass) and enc.date == date)
        target = next(enc_iter, None)
        if not target:
            # Make the new encounter.
            target = klass(date=date)
            # Add the new encounter to the subject encounters list.
            self._subject.add_encounter(target)

        # Return the target encounter.
        return target

    def encounter_type(self, row):
        """
        Infers the encounter type from the given row. This base
        implementation returns the parsed row *intervention_type*
        value.

        :param row: the input row
        :return: the REST data model Encounter subclass
        """
        return row.intervention_type



import os
import shutil
from nipype.interfaces.base import (traits, BaseInterfaceInputSpec,
                                    TraitedSpec, BaseInterface,
                                    File, Directory)


class MoveInputSpec(BaseInterfaceInputSpec):
    in_file = traits.Either(File, Directory, exists=True,
                            mandatory=True, desc='The file or directory to move')

    dest = Directory(mandatory=True, desc='The destination path')


class MoveOutputSpec(TraitedSpec):
    out_file = traits.Either(
        File, Directory, exists=True, desc='The moved file or directory')


class Move(BaseInterface):
    """The Move interface moves a file to a destination directory."""
    
    input_spec = MoveInputSpec

    output_spec = MoveOutputSpec

    def _run_interface(self, runtime):
        self.out_file = self._move(self.inputs.in_file, self.inputs.dest)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self.out_file
        return outputs

    def _move(self, in_file, dest):
        """
        Moves the given file.
    
        :param in_file: the path of the file to move
        @parma dest: the destination directory path
        :return: the moved file path
        """
        dest = os.path.abspath(dest)
        if not os.path.exists(dest):
            os.makedirs(dest)
        shutil.move(in_file, dest)
        _, fname = os.path.split(in_file)
        out_file = os.path.join(dest, fname)
        return out_file



import os
import re
import tempfile
import logging
from nipype.pipeline import engine as pe
from nipype.interfaces.utility import (IdentityInterface, Function, Merge)
from nipype.interfaces.ants import (AverageImages, Registration,
                                    ApplyTransforms)
from nipype.interfaces import fsl
from nipype.interfaces.dcmstack import CopyMeta
import qiutil
from qiutil.logging import logger
from ..interfaces import (Copy, XNATUpload)
from ..helpers import bolus_arrival
from .workflow_base import WorkflowBase
from .pipeline_error import PipelineError

REG_PREFIX = 'reg'
"""The XNAT registration resource name prefix."""

TECHNIQUES = ['ants', 'fnirt', 'mock']
"""The built-in registration techniques."""


def run(technique, project, subject, session, scan, ref_0, *images, **opts):
    """
    Runs the registration workflow on the given session scan images.

    :param technique: the :attr:`RegistrationWorkflow.technique`
    :param project: the project name
    :param subject: the subject name
    :param session: the session name
    :param scan: the scan number
    :param ref_0: the initial fixed reference image
    :param images: the image files to register
    :param opts: the :class:`RegistrationWorkflow` initializer
        and :meth:`RegistrationWorkflow.run` options
    :return: the realigned image file path array
    """
    # Extract the run options.
    run_opts = {k: opts.pop(k) for k in ['dest', 'mask'] if k in opts}
    # Make the workflow.
    reg_wf = RegistrationWorkflow(project=project, technique=technique,
                                  **opts)
    # Run the workflow.
    return reg_wf.run(subject, session, scan, ref_0, *images, **run_opts)


def generate_resource_name(technique):
    """
    Makes a unique registration resource name. Uniqueness permits more
    than one registration to be stored for a given session without a
    name conflict.

    :param technique: the registration technique
    :return: a unique XNAT registration resource name
    """
    return "%s_%s_%s" % (REG_PREFIX, technique,
                         qiutil.file.generate_file_name())


class RegistrationWorkflow(WorkflowBase):
    """
    The RegistrationWorkflow class builds and executes the registration workflow.
    The workflow registers an input NIfTI scan image against the input reference
    image and uploads the realigned image to XNAT.

    The registration workflow input is the *input_spec* node consisting of the
    following input fields:

    - *subject*: the subject name

    - *session*: the session name

    - *scan*: the scan number

    - *mask*: the mask to apply to the images

    - *reference*: the fixed reference image

    - *image*: the image file to register

    The mask can be obtained by running the
    :class:`qipipe.pipeline.mask.MaskWorkflow` workflow.

    The reference can be obtained by running the
    :class:`qipipe.pipeline.reference.ReferenceWorkflow` workflow.

    The output realigned image file is named the same as the input scan
    image file name is in the XNAT registration resource. The default
    resource name is auto-generated.
    
    Three registration techniques are supported:

    - ``ants``: ANTS_ SyN_ symmetric normalization diffeomorphic registration
      (default)

    - ``fnirt``: FSL_ FNIRT_ non-linear registration
    
    - ``mock``: Test technique which copies each input scan image to the
      output image file

    The optional workflow configuration file can contain overrides for the
    Nipype interface inputs in the following sections:

    - ``ants.Registration``: the ANTS `Registration interface`_ options

    - ``ants.ApplyTransforms``: the ANTS `ApplyTransform interface`_ options

    - ``fsl.FNIRT``: the FSL `FNIRT interface`_ options

    :Note: Since the XNAT *resource* name is unique, a
        :class:`qipipe.pipeline.registration.RegistrationWorkflow`
        instance can be used for only one registration workflow.
        Different registration inputs require different
        :class:`qipipe.pipeline.registration.RegistrationWorkflow`
        instances.

    .. _ANTS: http://stnava.github.io/ANTs/
    .. _ApplyTransform interface: http://nipy.sourceforge.net/nipype/interfaces/generated/nipype.interfaces.ants.resampling.html
    .. _FNIRT: http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FNIRT#Research_Overview
    .. _FNIRT interface: http://nipy.sourceforge.net/nipype/interfaces/generated/nipype.interfaces.fsl.preprocess.html
    .. _FSL: http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FSL
    .. _Registration interface: http://nipy.sourceforge.net/nipype/interfaces/generated/nipype.interfaces.ants.registration.html
    .. _SyN: http://www.ncbi.nlm.nih.gov/pubmed/17659998
    """

    def __init__(self, **kwargs):
        """
        If the optional configuration file is specified, then the workflow
        settings in that file override the default settings.

        :param kwargs: the :class:`qipipe.pipeline.workflow_base.WorkflowBase`
            initializer options, as well as the following options:
        :keyword resource: the XNAT resource name to use
        :keyword recursive: flag indicating whether to perform step-wise
            iterative recursive registration
        """
        super(RegistrationWorkflow, self).__init__(logger=logger(__name__),
                                                   **kwargs)

        technique_opt = kwargs.pop('technique', None)
        if not technique_opt:
            raise PipelineError('The registration technique was not specified.')
        self.technique = technique_opt.lower()
        """
        The lower-case XNAT registration technique. The built-in techniques
        include ``ants``, `fnirt`` and ``mock``.
        """

        rsc_opt = kwargs.pop('resource', None)
        self.resource = rsc_opt or generate_resource_name(self.technique)
        """The XNAT resource name used for all runs against this
        workflow instance."""

        self.workflow = self._create_realignment_workflow(**kwargs)
        """The registration realignment workflow."""

    def run(self, subject, session, scan, ref_0, *images, **opts):
        """
        Runs the registration workflow on the given session scan images.

        :param subject: the subject name
        :param session: the session name
        :param scan: the scan number
        :param ref_0: the required bolus uptake volume index
        :param images: the input session scan images
        :param opts: the following options:
        :option mask: the image mask file path
        :option dest: the realigned image target directory (default is the
            current directory)
        :return: the realigned output file paths
        """
        if not images:
            return []
        # Sort the images by volume number.
        sorted_scans = sorted(images)

        # The target location.
        dest = opts.get('dest')
        if dest:
            dest = os.path.abspath(dest)
        else:
            dest = os.getcwd()

        # The recursive flag.
        recursive = opts.get('recursive', False)
        # The execution workflow.
        exec_wf = self._create_execution_workflow(ref_0, dest, recursive)

        # Set the execution workflow inputs.
        input_spec = exec_wf.get_node('input_spec')
        input_spec.inputs.subject = subject
        input_spec.inputs.session = session
        input_spec.inputs.scan = scan
        mask = opts.get('mask')
        if mask:
            input_spec.inputs.mask = mask

        # Iterate over the registration inputs.
        iter_reg_input = exec_wf.get_node('iter_reg_input')
        iter_reg_input.iterables = ('image', images)

        # Execute the workflow.
        self.logger.debug("Executing the %s workflow on %s %s..." %
                         (self.workflow.name, subject, session))
        self._run_workflow(exec_wf)
        self.logger.debug("Executed the %s workflow on %s %s." %
                         (self.workflow.name, subject, session))

        # Return the output files.
        return [os.path.join(dest, filename(image)) for image in images]

    def _create_execution_workflow(self, ref_0, dest, recursive=False):
        """
        Makes the registration execution workflow on the given session
        scan images.

        The execution workflow input is the *input_spec* node consisting of
        the following input fields:

        - *subject*: the subject name

        - *session*: the session name

        - *scan*: the scan number

        - *mask*: the mask to apply to the images

        - *reference*: the fixed reference for the given image registration
        
        In addition, the caller has the responsibility of setting the
        ``iter_reg_input`` iterables to the 3D image files to realign.

        The *reference* input is set by :meth:`recurse`.

        :param ref_0: the initial fixed reference image
        :param dest: the target realigned image directory
        :param recursive: whether to use each realigned imaged as the
            succeeding registration reference
        :return: the execution workflow
        """
        if not ref_0:
            raise PipelineError('Registration workflow is missing the initial' +
                                ' fixed reference image')
        if not dest:
            raise PipelineError('Registration workflow is missing the' +
                             ' destination directory')
        self.logger.debug("Creating the registration execution workflow"
                           " with initial reference %s..." % ref_0)

        # The execution workflow.
        exec_wf = pe.Workflow(name='reg_exec', base_dir=self.workflow.base_dir)

        # The registration workflow input.
        input_fields = ['subject', 'session', 'scan', 'mask', 'ref_0',
                        'resource']
        input_spec = pe.Node(IdentityInterface(fields=input_fields),
                             name='input_spec')
        # The image hierarchy.
        exec_wf.connect(input_spec, 'subject',
                        self.workflow, 'input_spec.subject')
        exec_wf.connect(input_spec, 'session',
                        self.workflow, 'input_spec.session')
        exec_wf.connect(input_spec, 'scan',
                        self.workflow, 'input_spec.scan')
        # The registration mask.
        exec_wf.connect(input_spec, 'mask',
                        self.workflow, 'input_spec.mask')
        # The initial fixed reference image.
        input_spec.inputs.ref_0 = ref_0
        # The registration resource name.
        input_spec.inputs.resource = self.resource

        # The realignment child workflow iterator.
        iter_reg_fields = ['image', 'reference']
        iter_reg_input = pe.Node(IdentityInterface(fields=iter_reg_fields),
                                 name='iter_reg_input')
        exec_wf.connect(iter_reg_input, 'image',
                        self.workflow, 'input_spec.moving_image')
        exec_wf.connect(iter_reg_input, 'reference',
                        self.workflow, 'input_spec.reference')

        # If the recursive flag is set, then set the recursive
        # realigned -> reference connections. Otherwise, the fixed
        # reference is always the initial ref_0 image.
        if recursive:
            exec_wf.connect_iterables(iter_reg_input, copy_realigned, 
                                      recurse, ref_0=ref_0)
        else:
            iter_reg_input.inputs.reference = ref_0

        # The output destination directory.
        if not os.path.exists(dest):
            os.makedirs(dest)

        # Collect the realigned images.
        collect_realigned = pe.JoinNode(IdentityInterface(fields=['images']),
                                        joinsource='iter_reg_input',
                                        joinfield='images',
                                        name='collect_realigned')
        exec_wf.connect(self.workflow, 'output_spec.out_file',
                        collect_realigned, 'images')

        # Upload the registration result into the XNAT registration
        # resource.
        upload_reg_xfc = XNATUpload(project=self.project,
                                    resource=self.resource,
                                    modality='MR')
        upload_reg = pe.Node(upload_reg_xfc, name='upload_reg')
        exec_wf.connect(input_spec, 'subject', upload_reg, 'subject')
        exec_wf.connect(input_spec, 'session', upload_reg, 'session')
        exec_wf.connect(input_spec, 'scan', upload_reg, 'scan')
        exec_wf.connect(collect_realigned, 'images', upload_reg, 'in_files')

        # Copy the realigned images to the destination directory.
        #
        # FIXME: copying the realigned images individually with a Copy
        # submits a separate SGE job for each copy, contrary to the
        # defaults.cfg Copy setting. Although the defaults.cfg is
        # read into the registration workflow config, the defaults
        # are not applied, as they are for the parent qipipeline
        # workflow and its directly wired child workflows. The
        # registration workflow is built indirectly via the master
        # qipipeline register function for each scan input.
        # The work-around is to copy all of the realigned files
        # without a SGE submit in the copy_files function.
        # TODO: why doesn't registration recognize the default.cfg
        # Copy run_without_submitting setting?
        copy_realigned_xfc = Function(input_names=['in_files', 'dest'],
                                      output_names=['out_files'],
                                      function=copy_files)
        copy_realigned = pe.Node(copy_realigned_xfc, name='copy_realigned')
        copy_realigned.inputs.dest = dest
        exec_wf.connect(collect_realigned, 'images', copy_realigned, 'in_files')

        # The execution output.
        output_spec = pe.Node(IdentityInterface(fields=['images']),
                              name='output_spec')
        exec_wf.connect(collect_realigned, 'images', output_spec, 'images')

        self.logger.debug("Created the %s workflow." % exec_wf.name)
        # If debug is set, then diagram the workflow graph.
        if self.logger.level <= logging.DEBUG:
            self.depict_workflow(exec_wf)

        return exec_wf

    def _create_realignment_workflow(self, **opts):
        """
        Creates the workflow which registers and resamples images.
        The registration workflow performs the following steps:

        - Generates a unique XNAT resource name

        - Set the mask and realign workflow inputs

        - Run these workflows

        - Upload the realign outputs to XNAT
        
        :param opts: the following workflow options:
        :keyword base_dir: the workflow directory
            (default a new temp directory)
        :return: the Workflow object
        """
        self.logger.debug('Creating the registration realignment workflow...')

        # The workflow.
        base_dir = opts.get('base_dir', tempfile.mkdtemp(prefix='qipipe_'))
        realign_wf = pe.Workflow(name='registration', base_dir=self.base_dir)

        # The workflow input.
        in_fields = ['subject', 'session', 'scan', 'moving_image', 'reference',
                     'mask', 'resource']
        input_spec = pe.Node(IdentityInterface(fields=in_fields),
                             name='input_spec')
        input_spec.inputs.resource = self.resource

        # Copy the DICOM meta-data. The copy target is set by the technique
        # node defined below.
        copy_meta = pe.Node(CopyMeta(), name='copy_meta')
        realign_wf.connect(input_spec, 'moving_image', copy_meta, 'src_file')

        # The input file name without directory.
        input_filename_xfc = Function(input_names=['in_file'],
                                      output_names=['out_file'],
                                      function=filename)
        input_filename = pe.Node(input_filename_xfc, name='input_filename')
        realign_wf.connect(input_spec, 'moving_image',
                           input_filename, 'in_file')

        if self.technique.lower() == 'ants':
            # Nipype bug work-around:
            # Setting the registration metric and metric_weight inputs
            # after the node is created results in a Nipype input trait
            # dependency warning. Avoid this warning by setting these
            # inputs in the constructor from the values in the configuration.
            reg_cfg = self._interface_configuration(Registration)
            metric_inputs = {field: reg_cfg[field]
                             for field in ['metric', 'metric_weight']
                             if field in reg_cfg}
            # Register the images to create the rigid, affine and SyN
            # ANTS transformations.
            register = pe.Node(Registration(**metric_inputs), name='register')
            realign_wf.connect(input_spec, 'reference',
                               register, 'fixed_image')
            realign_wf.connect(input_spec, 'moving_image',
                               register, 'moving_image')
            realign_wf.connect(input_spec, 'mask',
                               register, 'fixed_image_mask')
            realign_wf.connect(input_spec, 'mask',
                               register, 'moving_image_mask')
            # Apply the transforms to the input image.
            apply_xfm = pe.Node(ApplyTransforms(), name='apply_xfm')
            realign_wf.connect(input_spec, 'reference',
                               apply_xfm, 'reference_image')
            realign_wf.connect(input_spec, 'moving_image',
                               apply_xfm, 'input_image')
            realign_wf.connect(input_filename, 'out_file',
                               apply_xfm, 'output_image')
            realign_wf.connect(register, 'forward_transforms',
                               apply_xfm, 'transforms')
            # Copy the meta-data.
            realign_wf.connect(apply_xfm, 'output_image',
                               copy_meta, 'dest_file')
        elif self.technique.lower() == 'fnirt':
            # Make the affine transformation.
            flirt = pe.Node(fsl.FLIRT(), name='flirt')
            realign_wf.connect(input_spec, 'reference', flirt, 'reference')
            realign_wf.connect(input_spec, 'moving_image', flirt, 'in_file')
            # Copy the input to a work directory, since FNIRT adds
            # temporary files to the input image location.
            fnirt_copy_moving = pe.Node(Copy(), name='fnirt_copy_moving')
            realign_wf.connect(input_spec, 'moving_image',
                               fnirt_copy_moving, 'in_file')
            # Register the image.
            fnirt = pe.Node(fsl.FNIRT(), name='fnirt')
            realign_wf.connect(input_spec, 'reference', fnirt, 'ref_file')
            realign_wf.connect(flirt, 'out_matrix_file', fnirt, 'affine_file')
            realign_wf.connect(fnirt_copy_moving, 'out_file', fnirt, 'in_file')
            realign_wf.connect(input_spec, 'mask', fnirt, 'inmask_file')
            realign_wf.connect(input_spec, 'mask', fnirt, 'refmask_file')
            realign_wf.connect(input_filename, 'out_file', fnirt, 'warped_file')
            # Copy the meta-data.
            realign_wf.connect(fnirt, 'warped_file', copy_meta, 'dest_file')
        elif self.technique.lower() == 'mock':
            # Copy the input scan file to an output file.
            mock_copy = pe.Node(Copy(), name='mock_copy')
            realign_wf.connect(input_spec, 'moving_image',
                               mock_copy, 'in_file')
            realign_wf.connect(mock_copy, 'out_file', copy_meta, 'dest_file')
        else:
            raise PipelineError("Registration technique not recognized: %s" %
                                self.technique)

        # The output is the realigned image.
        output_spec = pe.Node(IdentityInterface(fields=['out_file']),
                              name='output_spec')
        realign_wf.connect(copy_meta, 'dest_file', output_spec, 'out_file')

        self._configure_nodes(realign_wf)

        self.logger.debug("Created the %s workflow." % realign_wf.name)
        # If debug is set, then diagram the workflow graph.
        if self.logger.level <= logging.DEBUG:
            self.depict_workflow(realign_wf)
        return realign_wf


### Utility functions called by the workflow nodes. ###

def filename(in_file):
    """
    :param in_file: the input file path
    :return: the file name without a directory
    """
    import os

    return os.path.split(in_file)[1]


def copy_files(in_files, dest):
    """
    :param in_files: the input files
    :param dest: the destination directory
    :return: the output files
    """
    from qipipe.interfaces import Copy
    
    return [Copy(in_file=in_file, dest=dest).run().outputs.out_file
            for in_file in in_files]


def recurse(workflow, input_nodes, output_nodes, ref_0):
    """
    Connects the reference input for the given reference workflow input
    nodes. The reference for the first input node is the *ref_0* file.
    The reference for each subsequent node is the prior registration
    result.

    :param workflow: the workflow delegate which connects nodes
    :param input_nodes: the iterable expansion input scan image nodes
    :param output_nodes: the iterable expansion registration output
        nodes
    :param ref_0: the starting reference input node
    """
    # The reference for the first node is the initial reference.
    input_nodes[0].inputs.reference = ref_0
    # The reference for the remaining nodes is the previous registration
    # result.
    for i in range(1, node_cnt - 1):
        workflow.connect(output_nodes[i], 'out_file',
                         input_nodes[i + 1], 'reference')



import os
import shutil
from nipype.interfaces.base import (traits, BaseInterfaceInputSpec,
                                    TraitedSpec, BaseInterface,
                                    File, Directory)


class CopyInputSpec(BaseInterfaceInputSpec):
    in_file = traits.Either(File, Directory, exists=True, mandatory=True,
                            desc='The file or directory to copy')

    dest = Directory(desc='The destination directory path'
                          ' (default current directory)')

    out_fname = traits.Either(File, Directory,
                              desc='The destination file name'
                                   ' (default is the input file name)')


class CopyOutputSpec(TraitedSpec):
    out_file = traits.Either(File, Directory, exists=True,
                             desc='The copied file or directory')


class Copy(BaseInterface):
    """The Copy interface copies a file to a destination directory."""
    input_spec = CopyInputSpec

    output_spec = CopyOutputSpec

    def _run_interface(self, runtime):
        self._out_file = self._copy(self.inputs.in_file, self.inputs.dest,
                                    self.inputs.out_fname)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._out_file

        return outputs

    def _copy(self, in_file, dest=None, out_fname=None):
        """
        Copies the given file.
    
        :param in_file: the path of the file or directory to copy
        :param dest: the destination directory path
            (default is the current directory)
        :param out_fname: the destination file name
            (default is the input file name)
        :return: the copied file path
        """
        if dest:
            dest = os.path.abspath(dest)
            if not os.path.exists(dest):
                os.makedirs(dest)
        else:
            dest = os.getcwd()
        if out_fname:
            # Remove the out file name directory.
            _, out_fname = os.path.split(out_fname)
        else:
            # The default out file name is the input file name. 
            _, out_fname = os.path.split(in_file)
        out_file = os.path.join(dest, out_fname)
        if os.path.isdir(in_file):
            shutil.copytree(in_file, out_file)
        else:
            shutil.copy(in_file, out_file)
        
        return out_file



import os
import re
import glob
from setuptools import (setup, find_packages)

VCS_RQMT_PAT = re.compile('^\w+\+\w+:')
"""
The pattern for detecting a VCS requirement spec, e.g.
``git+git://...``.
"""


def version(package):
    """
    Return package version as listed in the `__init.py__` `__version__`
    variable.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def requires():
    """
    @return: the ``requirements.txt`` package specifications
    
    :Note: ``pip`` supports VCS package specifications, but
       setup.py does not. Therefore, this method filters out
       the VCS requirements in ``requirements.txt``. The VCS
       dependencies must be installed separately as described
       in the User Guide **Installation** section.
    """
    with open('requirements.txt') as f:
        rqmts = f.read().splitlines()
        return [rqmt for rqmt in rqmts if not VCS_RQMT_PAT.match(rqmt)]


def readme():
    with open("README.rst") as f:
        return f.read()


setup(
    name = 'qipipe',
    version = version('qipipe'),
    author = 'OHSU Knight Cancer Institute',
    author_email = 'loneyf@ohsu.edu',
    platforms = 'Any',
    license = 'MIT',
    keywords = 'Imaging QIN OHSU DCE MR XNAT Nipype',
    packages = find_packages(exclude=['test**']),
    package_data = dict(qipipe=['conf/*.cfg']),
    scripts = glob.glob('bin/*'),
    url = 'http://qipipe.readthedocs.org/en/latest/',
    description = 'Quantitative Imaging Profile pipeline',
    long_description = readme(),
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    install_requires = requires()
)



from nipype.interfaces.base import (traits,
                                    BaseInterfaceInputSpec, TraitedSpec, BaseInterface,
                                    InputMultiPath, OutputMultiPath, Directory, File)
from qidicom import hierarchy


class GroupDicomInputSpec(BaseInterfaceInputSpec):
    tag = traits.String(mandatory=True)
    
    in_files = InputMultiPath(
        traits.Either(File(exists=True), Directory(exists=True)),
        mandatory=True, desc='The DICOM files to group')


class GroupDicomOutputSpec(TraitedSpec):
    groups = traits.Dict(
        desc='The {tag value: [DICOM files]} dictionary'
    )


class GroupDicom(BaseInterface):
    input_spec = GroupDicomInputSpec

    output_spec = GroupDicomOutputSpec

    def _run_interface(self, runtime):
        self.grp_dict = hierarchy.group_by(self.inputs.tag, *self.inputs.in_files)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['groups'] = self.grp_dict
        return outputs



from nipype.interfaces.base import (
    traits, BaseInterfaceInputSpec, TraitedSpec,
    BaseInterface, File, Directory)
import qixnat

CONTAINER_OPTS = ['container_type', 'scan', 'reconstruction', 'assessor']
"""The download input container options."""


class XNATDownloadInputSpec(BaseInterfaceInputSpec):
    project = traits.Str(mandatory=True, desc='The XNAT project id')

    subject = traits.Str(mandatory=True, desc='The XNAT subject name')

    session = traits.Str(mandatory=True, desc='The XNAT session name')

    scan = traits.Either(traits.Str, traits.Int,
                         desc='The XNAT scan label or number')

    reconstruction = traits.Str(desc='The XNAT reconstruction name')

    assessor = traits.Str(desc='The XNAT assessor name')

    resource = traits.Str(desc='The XNAT resource name')

    container_type = traits.Enum('scan', 'reconstruction', 'assessor',
                                 desc='The XNAT resource container type')

    inout = traits.Str(desc='The XNAT resource in/out designator')

    file = traits.Str(desc="The XNAT file name (default all resource files)")

    dest = Directory(desc='The download location')


class XNATDownloadOutputSpec(TraitedSpec):
    out_files = traits.List(File(exists=True), desc='The downloaded files')
    out_file = File(exists=True,
                    desc='The downloaded file, if exactly one file was downloaded')


class XNATDownload(BaseInterface):
    """
    The ``XNATDownload`` Nipype interface wraps the
    :meth:`qixnat.facade.XNAT.download` method.
    
    :Note: only one XNAT operation can run at a time.
    
    Examples:
    
    >>> # Download the scan NIfTI files.
    >>> from qipipe.interfaces import XNATDownload
    >>> XNATDownload(project='QIN', subject='Breast003',
    ...     session='Session02', scan=1, resource='NIFTI',
    ...     dest='data').run()
    
    >>> # Download the scan DICOM files.
    >>> from qipipe.interfaces import XNATDownload
    >>> XNATDownload(project='QIN', subject='Breast003',
    ...     session='Session02', scan=1, resource='DICOM',
    ...     dest='data').run()
    
    >>> # Download the registration reg_H3pIz4s images.
    >>> from qipipe.interfaces import XNATDownload
    >>> XNATDownload(project='QIN', subject='Breast003',
    ...     session='Session02', resource='reg_H3pIz4',
    ...     dest='data').run()
    """

    input_spec = XNATDownloadInputSpec

    output_spec = XNATDownloadOutputSpec

    def _run_interface(self, runtime):
        opts = {}
        for ctr_type in CONTAINER_OPTS:
            ctr_name = getattr(self.inputs, ctr_type)
            if ctr_name:
                opts[ctr_type] = ctr_name
                break
        if self.inputs.dest:
            opts['dest'] = self.inputs.dest
        if self.inputs.resource:
            opts['resource'] = self.inputs.resource
        if self.inputs.file:
            opts['file'] = self.inputs.file
        with qixnat.connect() as xnat:
            self._out_files = xnat.download(self.inputs.project,
                                            self.inputs.subject,
                                            self.inputs.session, **opts)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_files'] = self._out_files
        if len(self._out_files) == 1:
            outputs['out_file'] = self._out_files[0]

        return outputs



import os
import re
from matplotlib import (pyplot, colors, cm)
import nibabel as nib
from qiutil.file import splitexts
from . import image


def create_lookup_table(ncolors, colormap='jet', out_file=None):
    """
    Generates a colormap lookup table with the given number of colors.
    
    :param ncolors: the number of colors to generate
    :param colormap: the matplotlib colormap name
    :param out_file: the output file path (default is the colormap
      name followed by ``_colors.txt``in the current directory)
    """
    # If there is an output file argument, then ensure that there is an
    # output parent directory. Otherwise, use the default output file.
    if out_file:
        out_dir = os.path.dirname(out_file)
        os.makedirs(out_dir)
    else:
        out_file = colormap + '_colors.txt'

    # The LUT reference values.
    values = range(ncolors)
    # The color map.
    cmap = pyplot.get_cmap(colormap)
    # The color range normalizer.
    norm = colors.Normalize(vmin=0, vmax=values[-1])
    # The value -> color mapper.
    scalar_map = cm.ScalarMappable(norm=norm, cmap=cmap)
    # The RGBA analog of the values.
    rgbas = scalar_map.to_rgba(values, bytes=True)

    # Save the LUT.
    print ("Saving the colormap as %s..." % out_file)
    with open(out_file, 'w') as f:
        # Write the initial background color.
        # Note - the LUT field separator must be one blank;
        # white space, e.g. a tab, is not supported by some
        # renderers, e.g. XTK.
        line = '0 0 0 0 0 0\n'
        f.write(line)
        # Write the colormap colors.
        for i, rgba in enumerate(rgbas):
            r, g, b, a = rgba
            line = "%d %d %d %d %d %d\n" % (i + 1, i + 1, r, g, b, a)
            f.write(line)


def colorize(lut_file, *inputs, **opts):
    """
    Transforms each input voxel value to a color lookup table reference.

    The input voxel values are uniformly partitioned for the given colormap
    LUT. For example, if the voxel values range from 0.0 to 3.0, then the
    a voxel value of 0 transformed to the first LUT color, 3.0 is transformed
    to the last LUT color, and the intermediate values are transformed to
    intermediate colors.

    The voxel -> reference output file name appends ``_color`` to the
    input file basename and preserves the input file extensions, e.g.
    the input file ``k_trans_map.nii.gz`` is transformed to
    ``k_trans_map_color.nii.gz`` in the output directory.
    
    :param lut_file: the color lookup table file path
    :param inputs: the image files to transform
    :param opts: the following options:
    @option dest: the destination directory (default current working directory)
    @option threshold: the threshold in the range 0 to nvalues (default 0)
    """
    dest = opts.pop('dest', None)
    if dest:
        if not os.path.exists(dest):
            os.makedirs(dest)
    else:
        dest = os.getcwd()
    opts.update(_infer_range_parameters(lut_file))
    opts['normalizer'] = _normalize
    for in_file in inputs:
        _colorize(in_file, dest, **opts)


def _normalize(value, vmin, vspan):
    # Zero always maps to the translucence in the first colormap LUT entry.
    if value == 0:
        return 0
    else:
        return image.normalize(value, vmin, vspan)


def _colorize(in_file, dest, **opts):
    # Split up the input file path.
    _, in_fname = os.path.split(in_file)
    base, exts = splitexts(in_fname)
    # The output file.
    out_fname = base + '_color' + exts
    out_file = os.path.join(dest, out_fname)
    image.discretize(in_file, out_file, **opts)


def _infer_range_parameters(lut_file):
    with open(lut_file) as f:
        content = f.readlines()
        nvalues = len(content)
        start = min(_labels(content))
        
        return dict(nvalues=nvalues, start=start)


LABEL_REGEX = re.compile('\d+')


def _labels(content):
    """Label generator for the given colormap LUT content."""
    for line in content:
        match = LABEL_REGEX.match(line)
        # Could be a comment, so ignore if no match.
        if match:
            yield int(match.group(0))




import os
from six.moves.configparser import ConfigParser as Config
from six.moves.configparser import NoOptionError

CFG_FILE = os.path.abspath(
    os.path.join( os.path.dirname(__file__), '..', 'conf', 'sarcoma.cfg')
)
"""
The Sarcoma Tumor Location configuration file. This file contains
properties that associat the subject name to the location, e.g.::

    Sarcoma004 = SHOULDER

The value is the SNOMED anatomy term.
"""


class ConfigError(Exception):
    pass


def sarcoma_location(subject):
    """
    :param subject: the XNAT Subject ID
    :return: the subject tumor location
    """
    try:
      return sarcoma_config().get('Tumor Location', subject)
    except NoOptionError:
      raise ConfigError("Tumor location for subject %s was not found in the"
                        " sarcoma configuration file %s" % (subject, CFG_FILE))


def sarcoma_config():
    """
    :return: the sarcoma configuration
    :rtype: ConfigParser
    """
    # Read the configuration file on demand.
    if not hasattr(sarcoma_config, 'instance'):
        sarcoma_config.instance = Config()
        sarcoma_config.instance.read(CFG_FILE)

    return sarcoma_config.instance
